/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPHeader;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPLocationInfo;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPTransportFeature;

/**
 * Models an <code>UnderlyingProtocol</code> component which statically binds SOAP 1.2 to the
 * HTTP transmission protocol. Contains a subset of the HTTP properties specified in the WSDL
 * 2.0 HTTP binding extension which are applicable to the WSDL 2.0 SOAP binding extension when
 * binding uses HTTP as the underlying protocol.
 * <p>
 * Note that the following extension attributes, which annotate an endpoint element with
 * transport related data, MUST be converted by a compliant <code>WSDLFactory</code> to
 * transport policy and attached to <code>Endpoint</code> where it MAY be consumed by
 * {@link HTTPTransportFeature}, i.e. if feature enabled.
 * <p>
 * whttp:authenticationScheme
 * <p>
 * whttp:authenticationRealm
 */
public final class SOAP12HTTPProtocol extends SOAPHTTPProtocol {

    public static final String TYPE = "http://www.w3.org/2003/05/soap/bindings/HTTP/";

    protected Boolean useHTTPCookies;
    protected Map<QName, Map<String, HTTPHeader>> faultHTTPHeaders;
    protected Map<QName, Map<Direction, Set<HTTPHeader>>> messageReferenceHTTPHeaders;
    protected String httpContentEncodingDefault;
    protected String httpQueryParameterSeparatorDefault;
    protected Map<QName, String> operationHTTPContentEncodings;
    protected Map<QName, String> faultHTTPContentEncodings;
    protected Map<QName, Map<Direction, String>> messageReferenceHTTPContentEncodings;
    protected Map<QName, HTTPLocationInfo> httpLocationInfos;
    protected Map<QName, String> httpQueryParameterSeparators;

    /**
     * Constructor which generates an instance of this class with default values, i.e. which
     * MAY be appropriate when reading binding definition from a WSDL 1.1 description.
     */
    public SOAP12HTTPProtocol() {
        this(null, new HashMap<QName, Map<String, HTTPHeader>>(),
                new HashMap<QName, Map<Direction, Set<HTTPHeader>>>(), null, null, new HashMap<QName, String>(),
                new HashMap<QName, String>(), new HashMap<QName, Map<Direction, String>>(),
                new HashMap<QName, HTTPLocationInfo>(), new HashMap<QName, String>());
    }

    /**
     * Constructs a SOAP 1.2 HTTP <code>UnderlyingProtocol</code> component which binds SOAP
     * 1.2 to the HTTP transmission protocol.
     * 
     * @param useHTTPCookies
     *        True if HTTP cookies as defined by IETF RFC 2965, are used for some or for all
     *        operations of <ocde>Interface</code> bound by parent <code>Binding</code> are
     *        used. False if cookies are not used. Null if no assertion is made regarding
     *        cookies.
     * @param faultHTTPHeaders
     *        Maps a fault QName to a map of HTTP Header entries, each of which maps a header
     *        name to an HTTPHeader. If no http headers are defined for a specific fault, then
     *        no entry exists. If no fault headers are defined, faultHTTPHeaders argument is an
     *        empty map.
     * @param messageReferenceHTTPHeaders
     *        A map whose entries map an operation name to a map of http headers. Each set of
     *        http headers is indexed by a Direction attribute which indicates whether
     *        associated messase is an input message or an output message. If an entry for an
     *        operation name exists, it MUST be mapped to a non-empty map of header entries. If
     *        no message headers are defined for normal messages on any operation,
     *        messageReferenceHTTPHeaders argument is an empty map.
     * @param httpContentEncodingDefault
     *        Indicates default transfer codings available for all <code>InterfaceMessageReference</code>
     *        and <code>InterfaceFaultReference</code> components of any <code>Interface</code>
     *        component that uses parent <code>Binding</code>, or <code>NULL</code> if no
     *        assertion is made regarding default transfer coding.
     * @param httpQueryParameterSeparatorDefault
     *        indicates the default query parameter separator character for all the <code>InterfaceOperation</code>
     *        components of any <code>Interface</code> component to which parent <code>Binding</code>
     *        is applied to.
     * @param operationHTTPContentEncodings
     *        Maps a <code>BindingOperation</code> name to default transfer codings available
     *        for all <code>BindingMessageReference</code> and <code>BindingFaultReference</code>
     *        components of indicated operation, or an empty collection if no codings defined.
     * @param faultHTTPContentEncodings
     *        Maps a <code>BindingFault</code> name to default transfer codings available to
     *        indicated fault, or an empty collection if no codings defined.
     * @param messageReferenceHTTPContentEncodings
     *        A map whose entries map an operation name to a map of transfer codings. Each
     *        transfer coding is indexed by a Direction attribute which indicates whether
     *        associated message is an input message or an output message. If an entry for an
     *        operation name exists, it MUST be mapped to a non-empty map of transfer coding
     *        entries. If no transfer codings are defined for normal messages on any operation,
     *        messageReferenceHTTPContentEncodings argument is an empty map.
     * @param httpLocationInfos
     *        A map whose entries map an operation name to an <code>HTTPLocationInfo</code>
     *        object, or an empty map if no location info defined.
     * @param httpQueryParameterSeparators
     *        A map whose entries map an operation name to a query parameter separator
     *        character, or an empty map if none defined.
     * 
     * @see http://www.w3.org/TR/soap12-part2
     * @see http://www.w3.org/TR/wsdl20-adjuncts/#soap-binding
     * 
     */

    public SOAP12HTTPProtocol(Boolean useHTTPCookies, Map<QName, Map<String, HTTPHeader>> faultHTTPHeaders,
            Map<QName, Map<Direction, Set<HTTPHeader>>> messageReferenceHTTPHeaders,
            String httpContentEncodingDefault, String httpQueryParameterSeparatorDefault,
            Map<QName, String> operationHTTPContentEncodings, Map<QName, String> faultHTTPContentEncodings,
            Map<QName, Map<Direction, String>> messageReferenceHTTPContentEncodings,
            Map<QName, HTTPLocationInfo> httpLocationInfos, Map<QName, String> httpQueryParameterSeparators) {
        super(TYPE);
        this.useHTTPCookies = useHTTPCookies;
        if (faultHTTPHeaders == null) {
            throw new IllegalArgumentException("faultHTTPHeaders null");
        } else {
            this.faultHTTPHeaders = faultHTTPHeaders;
        }
        if (messageReferenceHTTPHeaders == null) {
            throw new IllegalArgumentException("messageReferenceHTTPHeaders null");
        } else {
            this.messageReferenceHTTPHeaders = messageReferenceHTTPHeaders;
        }
        this.httpContentEncodingDefault = httpContentEncodingDefault;
        this.httpQueryParameterSeparatorDefault = httpQueryParameterSeparatorDefault;
        if (operationHTTPContentEncodings == null) {
            throw new IllegalArgumentException("operationHTTPContentEncodings null");
        } else {
            this.operationHTTPContentEncodings = operationHTTPContentEncodings;
        }
        if (faultHTTPContentEncodings == null) {
            throw new IllegalArgumentException("faultHTTPContentEncodings null");
        } else {
            this.faultHTTPContentEncodings = faultHTTPContentEncodings;
        }
        if (messageReferenceHTTPContentEncodings == null) {
            throw new IllegalArgumentException("messageReferenceHTTPContentEncodings null");
        } else {
            this.messageReferenceHTTPContentEncodings = messageReferenceHTTPContentEncodings;
        }
        for (Map<String, HTTPHeader> headerMap : faultHTTPHeaders.values()) {
            for (HTTPHeader httpHeader : headerMap.values()) {
                ((SOAPHTTPHeader)httpHeader).setParent(this);
            }
        }
        for (Map<Direction, Set<HTTPHeader>> headerMap : messageReferenceHTTPHeaders.values()) {
            for (Set<HTTPHeader> headerSet : headerMap.values()) {
                for (HTTPHeader httpHeader : headerSet) {
                    ((SOAPHTTPHeader)httpHeader).setParent(this);
                }
            }
        }
        if (httpLocationInfos == null) {
            throw new IllegalArgumentException("httpLocationInfos null");
        } else {
            this.httpLocationInfos = httpLocationInfos;
        }
        if (httpQueryParameterSeparators == null) {
            throw new IllegalArgumentException("httpQueryParameterSeparators null");
        } else {
            this.httpQueryParameterSeparators = httpQueryParameterSeparators;
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getFaultHTTPHeaders(javax.xml.namespace.QName)
     */
    @SuppressWarnings("unchecked")
    public Collection<HTTPHeader> getFaultHTTPHeaders(QName faultName) {
        if (faultHTTPHeaders == null) {
            return Collections.EMPTY_LIST;
        }
        Map<String, HTTPHeader> faultHeaders = faultHTTPHeaders.get(faultName);
        if (faultHeaders != null) {
            return Collections.unmodifiableCollection(faultHeaders.values());
        }
        return Collections.EMPTY_SET;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getFaultHTTPContentEncoding(javax.xml.namespace.QName)
     */
    public String getFaultHTTPContentEncoding(QName faultName) {
        if (faultHTTPContentEncodings == null) {
            return httpContentEncodingDefault;
        }
        String answer = faultHTTPContentEncodings.get(faultName);
        if (answer == null) {
            answer = httpContentEncodingDefault;
        }
        return answer;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getMessageReferenceHTTPHeaders(javax.xml.namespace.QName,
     *      org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction)
     */
    @SuppressWarnings("unchecked")
    public Collection<HTTPHeader> getMessageReferenceHTTPHeaders(QName operationName, Direction direction) {
        if (messageReferenceHTTPHeaders == null) {
            return Collections.EMPTY_LIST;
        }
        Map<Direction, Set<HTTPHeader>> directionMap = messageReferenceHTTPHeaders.get(operationName);
        if (directionMap != null) {
            Set<HTTPHeader> headerSet = directionMap.get(direction);
            if (headerSet != null) {
                return Collections.unmodifiableCollection(headerSet);
            }
        }
        return Collections.EMPTY_SET;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getMessageReferenceHTTPContentEncoding(javax.xml.namespace.QName,
     *      org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction)
     */
    public String getMessageReferenceHTTPContentEncoding(QName operationName, Direction direction) {
        if (messageReferenceHTTPContentEncodings == null) {
            return httpContentEncodingDefault;
        }
        String answer = null;
        Map<Direction, String> directionMap = messageReferenceHTTPContentEncodings.get(operationName);
        if (directionMap != null) {
            answer = directionMap.get(direction);
        }
        if (answer == null) {
            answer = operationHTTPContentEncodings.get(operationName);
        }
        if (answer == null) {
            answer = httpContentEncodingDefault;
        }
        return answer;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#useHTTPCookies()
     */
    public Boolean useHTTPCookies() {
        return useHTTPCookies;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getHTTPLocationInfo(javax.xml.namespace.QName)
     */
    public HTTPLocationInfo getHTTPLocationInfo(QName operationName) {
        if (httpLocationInfos == null) {
            return null;
        }
        return httpLocationInfos.get(operationName);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol#getHTTPQueryParamaterSeparator(javax.xml.namespace.QName)
     */
    public String getHTTPQueryParameterSeparator(QName operationName) {
        if (httpQueryParameterSeparators == null) {
            return httpQueryParameterSeparatorDefault;
        }
        String answer = httpQueryParameterSeparators.get(operationName);
        if (answer == null) {
            answer = httpQueryParameterSeparatorDefault;
        }
        return answer;
    }

    protected void validate() throws DeploymentException {
        if (!(parent instanceof SOAP12Binding)) {
            throw new DeploymentException("Parent not instanceof " + SOAP12Binding.class);
        }
    }

    void establishReferences() throws DeploymentException {
        if (faultHTTPHeaders != null) {
            for (Map<String, HTTPHeader> headerMap : faultHTTPHeaders.values()) {
                for (HTTPHeader httpHeader : headerMap.values()) {
                    ((SOAPHTTPHeader)httpHeader).establishReference();
                }
            }
        }
        if (messageReferenceHTTPHeaders != null) {
            for (Map<Direction, Set<HTTPHeader>> headerMap : messageReferenceHTTPHeaders.values()) {
                for (Set<HTTPHeader> headerSet : headerMap.values()) {
                    for (HTTPHeader httpHeader : headerSet) {
                        ((SOAPHTTPHeader)httpHeader).establishReference();
                    }
                }
            }
        }
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("type: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(TYPE);
        sb.append(System.getProperty("line.separator"));
        sb.append("useHTTPCookies: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(useHTTPCookies);
        sb.append(System.getProperty("line.separator"));
        sb.append("faultHTTPHeaders:");
        sb.append(System.getProperty("line.separator"));
        sb.append(faultHTTPHeaders);
        sb.append(System.getProperty("line.separator"));
        sb.append("messageReferenceHTTPHeaders:");
        sb.append(System.getProperty("line.separator"));
        sb.append(messageReferenceHTTPHeaders);
        sb.append(System.getProperty("line.separator"));
        sb.append("httpContentEncodingDefault:");
        sb.append(System.getProperty("line.separator"));
        sb.append(httpContentEncodingDefault);
        sb.append(System.getProperty("line.separator"));
        sb.append("httpQueryParameterSeparatorDefault:");
        sb.append(System.getProperty("line.separator"));
        sb.append(httpQueryParameterSeparatorDefault);
        sb.append(System.getProperty("line.separator"));
        sb.append("operationHTTPContentEncodings:");
        sb.append(System.getProperty("line.separator"));
        sb.append(operationHTTPContentEncodings);
        sb.append(System.getProperty("line.separator"));
        sb.append("faultHTTPContentEncodings:");
        sb.append(System.getProperty("line.separator"));
        sb.append(faultHTTPContentEncodings);
        sb.append(System.getProperty("line.separator"));
        sb.append("messageReferenceHTTPContentEncodings:");
        sb.append(System.getProperty("line.separator"));
        sb.append(messageReferenceHTTPContentEncodings);
        sb.append(System.getProperty("line.separator"));
        sb.append("httpLocationInfos:");
        sb.append(System.getProperty("line.separator"));
        sb.append(httpLocationInfos);
        sb.append(System.getProperty("line.separator"));
        sb.append("httpQueryParameterSeparators:");
        sb.append(System.getProperty("line.separator"));
        sb.append(httpQueryParameterSeparators);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(TYPE)
                + hashCode(useHTTPCookies)
                + hashCode(faultHTTPHeaders)
                + hashCode(messageReferenceHTTPHeaders)
                + hashCode(httpContentEncodingDefault)
                + hashCode(httpQueryParameterSeparatorDefault)
                + hashCode(operationHTTPContentEncodings)
                + hashCode(faultHTTPContentEncodings)
                + hashCode(messageReferenceHTTPContentEncodings)
                + hashCode(httpLocationInfos)
                + hashCode(httpQueryParameterSeparators);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof SOAP12HTTPProtocol)) {
            return false;
        }

        if ((useHTTPCookies == ((SOAP12HTTPProtocol)obj).useHTTPCookies || (useHTTPCookies != null && useHTTPCookies
                .equals(((SOAP12HTTPProtocol)obj).useHTTPCookies)))
                && (faultHTTPHeaders == ((SOAP12HTTPProtocol)obj).faultHTTPHeaders || (faultHTTPHeaders != null && faultHTTPHeaders
                        .equals(((SOAP12HTTPProtocol)obj).faultHTTPHeaders)))
                && (messageReferenceHTTPHeaders == ((SOAP12HTTPProtocol)obj).messageReferenceHTTPHeaders || (messageReferenceHTTPHeaders != null && messageReferenceHTTPHeaders
                        .equals(((SOAP12HTTPProtocol)obj).messageReferenceHTTPHeaders)))
                && (httpContentEncodingDefault == ((SOAP12HTTPProtocol)obj).httpContentEncodingDefault || (httpContentEncodingDefault != null && httpContentEncodingDefault
                        .equals(((SOAP12HTTPProtocol)obj).httpContentEncodingDefault)))
                && (httpQueryParameterSeparatorDefault == ((SOAP12HTTPProtocol)obj).httpQueryParameterSeparatorDefault || (httpQueryParameterSeparatorDefault != null && httpQueryParameterSeparatorDefault
                        .equals(((SOAP12HTTPProtocol)obj).httpQueryParameterSeparatorDefault)))
                && (operationHTTPContentEncodings == ((SOAP12HTTPProtocol)obj).operationHTTPContentEncodings || (operationHTTPContentEncodings != null && operationHTTPContentEncodings
                        .equals(((SOAP12HTTPProtocol)obj).operationHTTPContentEncodings)))
                && (faultHTTPContentEncodings == ((SOAP12HTTPProtocol)obj).faultHTTPContentEncodings || (faultHTTPContentEncodings != null && faultHTTPContentEncodings
                        .equals(((SOAP12HTTPProtocol)obj).faultHTTPContentEncodings)))
                && (messageReferenceHTTPContentEncodings == ((SOAP12HTTPProtocol)obj).messageReferenceHTTPContentEncodings || (messageReferenceHTTPContentEncodings != null && messageReferenceHTTPContentEncodings
                        .equals(((SOAP12HTTPProtocol)obj).messageReferenceHTTPContentEncodings)))
                && (httpLocationInfos == ((SOAP12HTTPProtocol)obj).httpLocationInfos || (httpLocationInfos != null && httpLocationInfos
                        .equals(((SOAP12HTTPProtocol)obj).httpLocationInfos)))
                && (httpQueryParameterSeparators == ((SOAP12HTTPProtocol)obj).httpQueryParameterSeparators || (httpQueryParameterSeparators != null && httpQueryParameterSeparators
                        .equals(((SOAP12HTTPProtocol)obj).httpQueryParameterSeparators)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
