/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import java.util.Collections;
import java.util.List;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.binding.BindingFault;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;

public final class SOAPBindingFault extends BindingFault {

    private SOAPMessage soapMessage;
    private QName soapFaultCode;
    private List<QName> soapFaultSubCodes;

    /**
     * Constructs a protocol specific, i.e. SOAP, <code>BindingFault</code> component.
     * 
     * @param abstractFaultName
     *        Qualified name of abstract fault within Interface bound by parent Binding to
     *        which this component is bound.
     * @param elementPolicy
     *        Contains policy expressions which may be attached to <code>BindingFault</code>.
     * @param soapMessage
     *        metadata which abstractly describes concrete message format
     * @param soapFaultCode
     *        Identifies a possible soap fault code for the operations in scope. If undefined,
     *        <code>null</code> is passed in which case the token '#any' is assumed, i.e. no
     *        assertion is made about the value of the SOAP fault code.
     * @param soapFaultSubCodes
     *        Identifies one or more subcodes for this SOAP fault. If undefined, an empty Set
     *        is passed in which case the token '#any' is assumed, i.e. no assertion is made
     *        regarding the value of SOAP fault subcodes.
     *        <p>
     *        If {@link SOAPVersion} employed by parent <code>Binding</code> is soap 1.1, no
     *        assertion is made about the value of SOAP fault subcodes, i.e. an empty set MUST
     *        be passed.
     */
    public SOAPBindingFault(QName abstractFaultName, ElementPolicy elementPolicy, SOAPMessage soapMessage,
            QName soapFaultCode, List<QName> soapFaultSubCodes) {
        super(abstractFaultName, elementPolicy);
        if (soapMessage == null) {
            throw new IllegalArgumentException("soapMessage null");
        } else {
            this.soapMessage = soapMessage;
        }
        this.soapFaultCode = soapFaultCode;
        if (soapFaultSubCodes == null) {
            throw new IllegalArgumentException("soapFaultSubCodes null");
        } else {
            this.soapFaultSubCodes = Collections.unmodifiableList(soapFaultSubCodes);
        }
        soapMessage.setParent(this);
    }

    /**
     * Returns qualified name of soap fault code.
     * @return qName or <code>null</code> in which case the token #any is assumed.
     */
    public QName getSoapFaultCode() {
        return soapFaultCode;
    }

    /**
     * Returns fault subcodes listed in increasing order of detail, i.e. if list contains more
     * than one subcode, the last code in the list MUST be the nested within the next to last
     * subcode progressing backwards through the list.
     * @return
     */
    @SuppressWarnings("unchecked")
    public List<QName> getSoapFaultSubCodes() {
        if (soapFaultSubCodes == null) {
            return Collections.EMPTY_LIST;
        }
        return soapFaultSubCodes;
    }

    public SOAPMessage getSoapMessage() {
        return soapMessage;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("soapMessage: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(soapMessage);
        sb.append(System.getProperty("line.separator"));
        sb.append("soapFaultCode: ");
        sb.append(soapFaultCode);
        sb.append(System.getProperty("line.separator"));
        sb.append("soapFaultSubCodes:");
        sb.append(System.getProperty("line.separator"));
        sb.append(soapFaultSubCodes);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(soapMessage) + hashCode(soapFaultCode) + hashCode(soapFaultSubCodes);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof SOAPBindingFault)) {
            return false;
        }

        if ((soapMessage == ((SOAPBindingFault)obj).soapMessage || (soapMessage != null && soapMessage
                .equals(((SOAPBindingFault)obj).soapMessage)))
                && (soapFaultCode == ((SOAPBindingFault)obj).soapFaultCode || (soapFaultCode != null && soapFaultCode
                        .equals(((SOAPBindingFault)obj).soapFaultCode)))
                && (soapFaultSubCodes == ((SOAPBindingFault)obj).soapFaultSubCodes || (soapFaultSubCodes != null && soapFaultSubCodes
                        .equals(((SOAPBindingFault)obj).soapFaultSubCodes)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
