/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.Component;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;

/**
 * Models a 'document style' <code>SOAPMessage</code>, i.e. a <code>SOAPMessage</code>
 * whose 'payload' is described by a 'GED', i.e. a global element declaration.
 */
@SuppressWarnings("unchecked")
public class SOAPDocumentStyleMessage extends NestedComponent implements SOAPMessage {

    private Set<QName> requiredHeaders;
    private Map<QName, SOAPHeaderBlock> headers;
    private String bodyPart;
    
    public SOAPDocumentStyleMessage(Map<QName, SOAPHeaderBlock> headers, String bodyPart) {
        if (headers == null) {
            throw new IllegalArgumentException("headers null");
        } else {
            this.headers = headers;
        }
        this.bodyPart = bodyPart;
        for (SOAPHeaderBlock header : headers.values()) {
            header.setParent(this);
            if (header.isRequired()) {
                if (requiredHeaders == null) {
                    requiredHeaders = new HashSet<QName>();
                }
                requiredHeaders.add(header.getElementName());
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#getMessageStyle()
     */
    public SOAPMessageStyle getMessageStyle() {
        return SOAPMessageStyle.DOCUMENT_STYLE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#getHeaders()
     */
    public Collection<SOAPHeaderBlock> getHeaders() {
        return headers.values();
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#getHeader(javax.xml.namespace.QName)
     */
    public SOAPHeaderBlock getHeader(QName name) {
        return headers.get(name);
    }

    /**
     * Returns name of abstract part defined on bound {@link InterfaceMessage} which defines
     * soap:body content.
     * 
     * @return String part name or <code>null</code> if part name is undefined, i.e.
     *         soap:body may contain any element or soap:body is empty.
     */
    public String getBodyPart() {
        return bodyPart;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#getRequiredHeaders()
     */
    public Set<QName> getRequiredHeaders() {
        return requiredHeaders;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    @SuppressWarnings("unchecked")
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "soapMessage", parent.getFragmentIdentifier().getPointerPart());
        }
        return fragmentIdentifier;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#setParent(org.bluestemsoftware.specification.eoa.component.Component)
     */
    @SuppressWarnings("unchecked")
    public void setParent(Component parent) {
        this.parent = parent;
    }
    
    // *********************** Identity Methods ************************************
    
    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("headers: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(headers);
        sb.append(System.getProperty("line.separator"));
        sb.append("bodyPart:");
        sb.append(System.getProperty("line.separator"));
        sb.append(bodyPart);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(headers) + hashCode(bodyPart);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof SOAPDocumentStyleMessage)) {
            return false;
        }

        if (
                (headers == ((SOAPDocumentStyleMessage)obj).headers || (headers != null && headers.equals(
                ((SOAPDocumentStyleMessage)obj).headers)))
                &&
                (bodyPart == ((SOAPDocumentStyleMessage)obj).bodyPart || (bodyPart != null && bodyPart.equals(
                        ((SOAPDocumentStyleMessage)obj).bodyPart)))     
        ) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
