/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.types.ElementDeclaration;
import org.bluestemsoftware.specification.eoa.component.types.Schema;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent.Type;

/**
 * Models a declared SOAP header by describing header data abstractly. A SOAP header is
 * contained within a {@link SOAPMessage}.
 */
public final class SOAPHeaderBlock extends NestedComponent<Binding, SOAPMessage> implements SchemaComponentReference {

    private QName elementName;
    private ElementDeclaration elementDecl;
    private boolean mustUnderstand;
    private boolean isRequired;
    private String partName;
    private Schema schema;

    /**
     * Constructor for <code>SOAPHeaderBlock</code>.
     * @param elementName
     *        qualified name of element decl which describes header component. Value MUST be
     *        unique within context of parent component.
     * @param mustUnderstand
     *        If true, indicates that runtime SOAP header block MUST provide a mustUnderstand
     *        attribute information item with a value of "true" or "1" as per the SOAP
     *        specification.
     *        <p>
     *        If false runtime header attribute is optional. If specified SOAP header block
     *        MUST provide a mustUnderstand attribute information item with a value of "false"
     *        or "0" as per the SOAP specification.
     * @param isRequired
     *        Indicates if the SOAP header block is required. If the value is "true", then the
     *        SOAP header block must be included in the message. If it is "false", then the
     *        SOAP header block is optional.
     */
    public SOAPHeaderBlock(QName elementName, boolean mustUnderstand, boolean isRequired) {
        super();
        if (elementName == null) {
            throw new IllegalArgumentException("elementName null");
        } else {
            this.elementName = elementName;
        }
        this.mustUnderstand = mustUnderstand;
        this.isRequired = isRequired;
    }

    /**
     * Constructor for <code>SOAPHeaderBlock</code>. This constructor is provided for
     * backwards compatibility with WSDL 1.1.
     * <p>
     * Note that implied value for {@link #mustUnderstand} attribute is true, and implied value
     * for {@link #isRequired} is true, i.e. runtime SOAP header block MUST be included in
     * message and it MUST specifiy a mustUnderstand attribute with value set to "1".
     * 
     * @param elementName
     *        qualified name of element decl which describes header component. Value MUST be
     *        unique within context of parent component.
     * @param partName
     *        name of part on abstract message to which this soap header is 'mapped'.
     */
    public SOAPHeaderBlock(QName elementName, String partName) {
        super();
        if (elementName == null) {
            throw new IllegalArgumentException("elementName null");
        } else {
            this.elementName = elementName;
        }
        if (partName == null) {
            throw new IllegalArgumentException("partName null");
        } else {
            this.partName = partName;
        }
        mustUnderstand = true;
        isRequired = true;
    }

    /**
     * Returns qualified name of element decl which describes header component. Value MUST be
     * unique within context of parent component.
     * 
     * @return QName qualified name.
     */
    public QName getElementName() {
        return elementName;
    }

    /**
     * Returns referenced element declaration component.
     * 
     * @return ElementDeclaration or <code>null</code> if reference not yet established
     */
    public ElementDeclaration getElementDecl() {
        return elementDecl;
    }

    /**
     * If true, indicates that runtime SOAP header block MUST provide a mustUnderstand
     * attribute information item with a value of "true" or "1" as per the SOAP specification.
     * <p>
     * If false runtime header attribute is optional. If specified SOAP header block MUST
     * provide a mustUnderstand attribute information item with a value of "false" or "0" as
     * per the SOAP specification.
     * <p>
     * Note that if true, the elementName MUST allow it, otherwise the attribute is optional.
     * 
     * @return boolean
     */
    public boolean mustUnderstand() {
        return mustUnderstand;
    }

    /**
     * Indicates if the SOAP header block is required. If the value is "true", then the SOAP
     * header block must be included in the message. If it is "false", then the SOAP header
     * block is optional.
     * 
     * @return boolean
     */
    public boolean isRequired() {
        return isRequired;
    }

    /**
     * Returns name of part on abstract message to which this soap header is 'mapped'.
     * <p>
     * This attribute is provided for backwards compatibility with WSDL 1.1. If
     * <code>null</code>, header cannot be described via WSDL 1.1.
     * <p>
     * If specified, value retrieved via {@link #elementName()} MUST match type defined on
     * corresponding abstract part which MUST be typed as an element.
     * 
     * @return String abstract part name or <code>null</code> if undefined.
     */
    public String getPartName() {
        return partName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference$SchemaComponentReference#getSchemaComponentName()
     */
    public String getSchemaComponentName() {
        return elementName.getLocalPart();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference$SchemaComponentReference#getSchemaComponentType()
     */
    public Type getSchemaComponentType() {
        return Type.ELEMENT_DECLARATION;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<Schema> getReferencedComponentClass() {
        return Schema.class;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public Schema getReferencedComponent() {
        return schema;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return new ComponentName(elementName.getNamespaceURI(), "foo");
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.SCHEMA;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(tns=");
            pointerPart.append(elementName.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("tns:");
            pointerPart.append(elementName.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "soapHeaderBlock",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(SOAPMessage parent) {
        this.parent = parent;
    }

    public void establishReference() throws DeploymentException {
        if (elementDecl != null) {
            return;
        }
        schema = getRootComponent().getComponentContext().findReferencedComponent(this);
        if (schema == null) {
            throw new DeploymentException("ComponentReference '"
                    + fragmentIdentifier
                    + "' is invalid. Referenced elementDeclaration "
                    + elementName
                    + ", is undefined.");
        } else {
            elementDecl = (ElementDeclaration)schema.getSchemaComponent(Type.ELEMENT_DECLARATION, elementName
                    .getLocalPart());
        }
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("elementName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(elementName);
        sb.append(System.getProperty("line.separator"));
        sb.append("isRequired:");
        sb.append(System.getProperty("line.separator"));
        sb.append(isRequired);
        sb.append(System.getProperty("line.separator"));
        sb.append("mustUnderstand:");
        sb.append(System.getProperty("line.separator"));
        sb.append(mustUnderstand);
        sb.append(System.getProperty("line.separator"));
        sb.append("partName:");
        sb.append(System.getProperty("line.separator"));
        sb.append(partName);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(elementName) + hashCode(isRequired) + hashCode(mustUnderstand) + hashCode(partName);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof SOAPHeaderBlock)) {
            return false;
        }

        if ((elementName == ((SOAPHeaderBlock)obj).elementName || (elementName != null && elementName
                .equals(((SOAPHeaderBlock)obj).elementName)))
                && (isRequired == ((SOAPHeaderBlock)obj).isRequired)
                && (mustUnderstand == ((SOAPHeaderBlock)obj).mustUnderstand)
                && (partName == ((SOAPHeaderBlock)obj).partName || (partName != null && partName
                        .equals(((SOAPHeaderBlock)obj).partName)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }
    
}
