/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.Component;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;

/**
 * Models an 'rpc style' <code>SOAPMessage</code>. Provided for backwards compatability with
 * WSDL 1.1. If <code>BindingOperation</code> defines an 'rpc style' message, parent
 * <code>Binding</code> cannot be described via WSDL 2.0.
 * <p>
 * Note that use of 'rpc style' messages is discouraged and MAY NOT be supported by all
 * {@link SOAPBinding} implementations. The intent of this specification is to encourage
 * loosely coupled, coarse grained, i.e. document driven, applications.
 */
@SuppressWarnings("unchecked")
public final class SOAPRpcStyleMessage extends NestedComponent implements SOAPMessage {

    private Set<QName> requiredHeaders;
    private Map<QName, SOAPHeaderBlock> headers;
    private String soapBodyNamespace;
    private List<String> bodyParts;

    public SOAPRpcStyleMessage(Map<QName, SOAPHeaderBlock> headers, String soapBodyNamespace,
            List<String> bodyParts) {
        if (headers == null) {
            throw new IllegalArgumentException("headers null");
        } else {
            this.headers = headers;
        }
        if (soapBodyNamespace == null) {
            throw new IllegalArgumentException("soapBodyNamespace null");
        } else {
            this.soapBodyNamespace = soapBodyNamespace;
        }
        if (bodyParts == null) {
            throw new IllegalArgumentException("bodyParts null");
        } else {
            this.bodyParts = bodyParts;
        }
        for (SOAPHeaderBlock header : headers.values()) {
            header.setParent(this);
            if (header.isRequired()) {
                if (requiredHeaders == null) {
                    requiredHeaders = new HashSet<QName>();
                }
                requiredHeaders.add(header.getElementName());
            }
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#getMessageStyle()
     */
    public SOAPMessageStyle getMessageStyle() {
        return SOAPMessageStyle.RPC_STYLE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#getHeaders()
     */
    public Collection<SOAPHeaderBlock> getHeaders() {
        return headers.values();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#getHeader(javax.xml.namespace.QName)
     */
    public SOAPHeaderBlock getHeader(QName name) {
        return headers.get(name);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#getRequiredHeaders()
     */
    public Set<QName> getRequiredHeaders() {
        return requiredHeaders;
    }

    /**
     * Indicates namespace of 'wrapper' element, i.e. soap body child, which 'wraps' part
     * accessor elements.
     * 
     * @return soapBodyNamespace or <code>null</code> if parent <code>SOAPMessage</code> is
     *         not a WSDL 1.1 'RPC Style' message.
     */
    public String getSoapBodyNamespace() {
        return soapBodyNamespace;
    }

    /**
     * Returns part names defined on {@link InterfaceMessage} bound by parent
     * <code>SOAPMessage</code> which define content model of part accessor elements.
     * 
     * @return List of part names or an empty set if remote procedure signature is empty.
     */
    public List<String> getBodyParts() {
        if (bodyParts == null) {
            return Collections.EMPTY_LIST;
        }
        return bodyParts;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "soapMessage",
                    parent.getFragmentIdentifier().getPointerPart());
        }
        return fragmentIdentifier;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage#setParent(org.bluestemsoftware.specification.eoa.component.Component)
     */
    public void setParent(Component parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("soapBodyNamespace: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(soapBodyNamespace);
        sb.append(System.getProperty("line.separator"));
        sb.append("headers: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(headers);
        sb.append(System.getProperty("line.separator"));
        sb.append("bodyParts:");
        sb.append(System.getProperty("line.separator"));
        sb.append(bodyParts);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(soapBodyNamespace) + hashCode(headers) + hashCode(bodyParts);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof SOAPRpcStyleMessage)) {
            return false;
        }

        if ((soapBodyNamespace == ((SOAPRpcStyleMessage)obj).soapBodyNamespace || (soapBodyNamespace != null && soapBodyNamespace
                .equals(((SOAPRpcStyleMessage)obj).soapBodyNamespace)))
                && (headers == ((SOAPRpcStyleMessage)obj).headers || (headers != null && headers
                        .equals(((SOAPRpcStyleMessage)obj).headers)))
                && (bodyParts == ((SOAPRpcStyleMessage)obj).bodyParts || (bodyParts != null && bodyParts
                        .equals(((SOAPRpcStyleMessage)obj).bodyParts)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
