/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap.rt;

import java.io.File;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.BusinessFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.FaultContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerRequest;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerResponse;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.rt.HyperTextTransferProtocolRT;

/**
 * A runtime binding which binds SOAP messaging protocol to to HTTP at runtime.
 */
public abstract class SOAPHTTPBindingRT extends SOAPBindingRT implements HyperTextTransferProtocolRT {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends SOAPBindingRT.Provider {

        public void spi_sendAction(EndpointReference endpointReference, ActionContext actionContext) throws SystemFault;

        public void spi_receiveAction(HTTPServerRequest request, HTTPServerResponse response, Direction direction, EndpointReference endpointReference);

    }

    public SOAPHTTPBindingRT(BindingFactory factory, Provider provider, ComponentName componentName,
            String protocolType, File extensionEtcDir, File extensionVarDir) {
        super(factory, provider, componentName, protocolType, extensionEtcDir, extensionVarDir);
    }

    /**
     * Invoked by an instance of <code>HTTPTransportModule</code>. Receives action addressed
     * to referenced endpoint. If referenced Service upon which action is defined is 'my'
     * Service, action represents a request, i.e. message direction is IN. If referenced
     * Service upon which action is defined is a 'partner' Service, action represents a
     * response, i.e. message direction is OUT.
     * <p>
     * Note that this extension uses the SOAP 1.2 'state machine' terminology along with
     * requirements defined within version 1.2 of the basic profile to define the expected
     * behavior of a compliant binding (both SOAP 1.1 and SOAP 1.2) when receiving an action.
     * <p>
     * The action MUST be considered as 'accepted for processing' by this binding, i.e. the
     * 'responding soap node', once the binding transitions from the 'init' state to the
     * 'receiving' state. If the action is not 'accepted for processing', a 4xx HTTP status
     * code and an empty response body MUST be returned.
     * <p>
     * If the action is 'accepted for processing' and referenced Service upon which action is
     * defined is a 'partner' service, a 2xx HTTP status code along with an empty response body
     * MUST be returned unless binding fails during the 'receiving' state due to a soap version
     * mismatch fault or a soap must understand fault, in which case a 500 HTTP status code
     * MUST be returned along with the corresponding soap fault message in the response body.
     * Note that if this binding, i.e. the 'responding soap node', fails during the 'receiving'
     * state for any reason other than a soap version mismatch or soap must understand fault,
     * the error MUST NOT be communicated to the 'requesting soap node'. It SHOULD, however, be
     * logged.
     * <p>
     * If the action is 'accepted for processing' and referenced Service upon which action is
     * defined is 'my' service and the WSDL MEP is 'in-only', a 2xx HTTP status code along with
     * an empty response body MUST be returned. If this binding, i.e. the 'responding soap
     * node' fails during the 'receiving' state for any reason other than a soap version
     * mismatch or soap must understand fault while processing a WSDL 'in-only' MEP, the error
     * MUST NOT be communicated to the 'requesting soap node'. It SHOULD, however, be logged.
     * <p>
     * If the action is 'accepted for processing' and referenced Service upon which action is
     * defined is 'my' service and the WSDL MEP is 'in-out' or 'robust-in-only' and the
     * response EPR uses a 'non-anonymous' wsa:Address value, a 2xx HTTP status code along with
     * an empty body MUST be returned. The corresponding response, i.e. SOAP message, MUST be
     * returned over a separate HTTP connection.
     * <p>
     * If the action is 'accepted for processing' and referenced Service upon which action is
     * defined is 'my' service and the WSDL MEP is 'in-out' or 'robust-in-only ' and the
     * response EPR uses an 'anonymous' wsa:Address value, and this binding, i.e. the
     * 'responding soap node' transitions from the 'receiving' state to the 'receiving+sending'
     * or 'sending' state, the response, i.e. SOAP message MUST be returned within the HTTP
     * response body. If the SOAP message is a SOAP fault, the status code MUST be 500. If the
     * SOAP message is a normal response, i.e. the WSDL MEP is 'in-out', the status code MUST
     * be 2xx.
     * <p>
     * Note that if referenced Service upon which action is defined is 'my' service and the
     * WSDL MEP is 'robust-in-only' the referenced endpoint SHOULD NOT require only anonymous
     * responses, i.e. to prevent the situation where requesting SOAP Node blocks for an
     * anonymous fault response that MAY never be returned.
     * 
     * @param request
     * @param response
     * @param direction
     * @param endpointReference
     * 
     * @see http://www.w3.org/TR/2003/REC-soap12-part2-20030624/#http-msgexop
     * @see http://www.ws-i.org/Profiles/BasicProfile-1_2(WGAD).html
     */
    public void receiveAction(HTTPServerRequest request, HTTPServerResponse response, Direction direction, EndpointReference endpointReference) {
        if (provider == null) {
            throw new IllegalStateException("provider null");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_receiveAction(request, response, direction, endpointReference);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Sends action to referenced <code>Endpoint</code>. If referenced Service upon which
     * action is defined is a 'partner' Service, action represents a request, i.e. message
     * direction is IN. If referenced Service upon which action is defined is 'my' Service,
     * action represents a response, i.e. message direction is OUT and WSDL MEP is either
     * 'in-out' or 'robust-in-only.'
     * <p>
     * Note that if referenced Service upon which action is defined is 'my' Service and a SOAP
     * message is present within the HTTP response body, the response MUST be discarded by a
     * compliant Binding provider.
     * <p>
     * Note that if referenced Service upon which action is defined is a 'partner' Service and
     * the WSDL MEP is either 'in-out' or 'robust-in-only' and a SOAP message is present within
     * the HTTP response body, the response MUST be returned by Binding to engine via the
     * method
     * {@link #receiveAction(HttpServletRequest, HttpServletResponse, EndpointReference)}
     * unless response was expected to be returned non-anonymously, in which case the response
     * MUST be discarded by a compliant Binding provider unless response is a soap must
     * understand fault or a soap version mismatch fault (both of which are always returned
     * anonymously.)
     * <p>
     * Note that if referenced Service upon which action is defined is a 'partner' Service and
     * the WSDL MEP is 'in-only' and a SOAP message is present within the HTTP response body,
     * the response MUST be discarded by a compliant Binding provider.
     * <p>
     * Note that if referenced Service upon which action is defined is a 'partner' Service and
     * the WSDL MEP is 'robust-in-only', and no assertion is made by receiving node regarding
     * anonymous responses, the wsa:FaultTo SHOULD indicate a non-anonymous response address,
     * i.e. to prevent the situation where requesting SOAP Node blocks for an anonymous fault
     * response that MAY never be returned.
     * 
     * @param ActionContext
     *        Defines context within which action occurs. If associated message direction is
     *        IN, action MUST be an action declared on Interface implemented by 'partner'
     *        Service, i.e. manifested as an instance of {@link MessageContext}.
     *        <p>
     *        If associated message direction is OUT and action is an output message declared
     *        on Interface implemented by 'my' Service, action MUST be manifested as an
     *        instance of {@link MessageContext}.
     *        <p>
     *        If associated message direction is OUT and action is a fault message declared on
     *        Interface implemented by 'my' Service, action MUST be manifested as an instance
     *        of {@link FaultContext} which contains a {@link BusinessFault#getMessage()}.
     *        <p>
     *        If associated message direction is OUT and action is an undeclared fault message
     *        returned by 'my' service, action MUST be manifested as an instance of
     *        {@link FaultContext} which contains a {@link SystemFault#getMessage()}.
     * 
     * @throws SystemFault
     *         Thrown by <code>Binding</code> as an indication that action was NOT 'accepted
     *         for processing' by the 'receiver.' Note that this extension uses the SOAP 1.2
     *         'state machine' terminology as a reference in determining whether or not an
     *         action has been 'accepted for processing.'
     *         <p>
     *         An action MUST be considered by Binding provider as 'accepted for processing'
     *         once this binding, i.e. the 'requesting soap node', transitions from the
     *         'requesting' state to the 'sending+receiving' or 'receiving' state (see the
     *         rules defined above regarding handling of the response, if any.)
     *         <p>
     *         If this Binding, i.e. the 'requesting soap node' either fails to initiate the
     *         action or if a failure is encountered within the 'requesting' state and the
     *         binding transitions into the 'fail' state, a <code>SystemFault</code> MUST be
     *         thrown. This Binding, i.e. the 'requesting soap node', MUST transition from the
     *         'requesting' state to the 'fail' state if the 'responding soap node' returns an
     *         HTTP status line where the status code is 4xx. This binding, i.e. the
     *         'requesting soap node' MUST NOT transition from the 'requesting' state to the
     *         'fail' state for any other reason. Note that codes within the 3xx range MUST be
     *         automatically handled by a compliant binding, i.e. by attempting to re-send
     *         action to the re-directed url(s). If after an unspecified number of
     *         attempts/redirects, the send is unsuccessfull, the binding MUST treat the error
     *         as though it were a 4xx related error.
     *         <p>
     *         Note that if an unchecked exception is encountered by Binding before action was
     *         'accepted for processing' by 'responding soap node', the exception MUST be
     *         converted by Binding to a <code>SystemFault</code> and thrown. If unchecked
     *         exception occurs after the action was 'accepted for processing' by 'responding
     *         soap node', the exception MUST be converted to a <code>SystemFault</code> and
     *         communicated to engine via the method
     *         {@link #receiveAction(HttpServletRequest, HttpServletResponse, EndpointReference)}.
     *         <p>
     *         Note that a compliant <code>Binding</code> provider MUST NOT throw a
     *         <code>SystemFault</code> for any other reason other than those cited above.
     * 
     * @see #createAction(String, Message)
     * @see http://www.w3.org/TR/2003/REC-soap12-part2-20030624/#http-msgexop
     */
    @Override
    public void sendAction(EndpointReference endpointReference, ActionContext actionContext) throws SystemFault {
        if (provider == null) {
            return;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_sendAction(endpointReference, actionContext);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

}
