/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.engine.default10;

import java.io.File;
import java.lang.reflect.Constructor;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EngineRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EngineReference;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.ext.ExecutableException;
import org.bluestemsoftware.specification.eoa.ext.Extension;
import org.bluestemsoftware.specification.eoa.ext.ExtensionException;
import org.bluestemsoftware.specification.eoa.ext.application.default10.DefaultApplication;
import org.bluestemsoftware.specification.eoa.system.ManagementContext;
import org.w3c.dom.Document;

/**
 * A 'default' <code>Engine</code> implementation.
 */
public class DefaultEngine extends EngineRT {

    public static final String TYPE = "http://bluestemsoftware.org/specification/eoa/ext/engine/default/1.0";
    public static final String PROVIDER_SCHEMA_LOCATION = "classpath:///schema/http.bluestemsoftware.org.specification.eoa.ext.engine.default.1.0.xsd";
        
    public interface Provider extends EngineRT.Provider {
    }

    public DefaultEngine(DefaultEngineFactory factory, ComponentName name, Document configuration,
            Map<QName, EngineReference> engineReferences, File extensionEtcDir, File extensionVarDir) {
        super(factory, new ProviderDelegate(configuration), name, engineReferences, extensionEtcDir, extensionVarDir);
    }

    @Override
    public Set<String> getRequiredApplicationFeatures() {
        // TODO: pull these out of configuration. we can't delegate
        // to underlying provider, because init has not been
        // invoked  yet, i.e. this method is called from
        // within the installfeaturemodules subroutine within
        // deploy
        return new HashSet<String>();
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.RootExtension$AbstractRootExtension#getExtensionProvider()
     */
    @Override
    public final org.bluestemsoftware.specification.eoa.ext.Extension.Provider getExtensionProvider() {
        return ((ProviderDelegate)provider).delegate;
    }

    /*
     * we can't instantiate user defined delegate until class name, which is defined on
     * application implemented by engine is available, i.e. once consumer establishes reference
     * to application during deployment.
     */
    private static class ProviderDelegate implements DefaultEngine.Provider {

        private DefaultEngine.Provider delegate;
        private DefaultEngine consumer;
        private Document configuration;

        public ProviderDelegate(Document configuration) {
            this.configuration = configuration;
        }

        /*
         * (non-Javadoc)
         * @see org.bluestemsoftware.specification.eoa.component.engine.rt.EngineRT.Provider#spi_destroy()
         */
        public void spi_destroy() {
            delegate.spi_destroy();
        }
        
        /*
         * (non-Javadoc)
         * @see org.bluestemsoftware.specification.eoa.component.engine.rt.EngineRT.Provider#spi_init(java.util.Set)
         */
        public void spi_init(Set<ManagementContext> managementContexts) throws ExtensionException {

            DefaultApplication app = (DefaultApplication)consumer.getApplication();
            String className = app.getClassName();
            if (className == null || className.equals("")) {
                throw new DefaultEngineException("Class name defined by application "
                        + app.getName()
                        + " is null or empty string");
            }
            ClassLoader cl = Thread.currentThread().getContextClassLoader();
            Class<?> clazz;
            try {
                clazz = cl.loadClass(className);
            } catch (ClassNotFoundException ce) {
                throw new DefaultEngineException("User defined class "
                        + className
                        + ", is not visible to classloader "
                        + cl.toString());
            }
            if (DefaultEngine.Provider.class.isAssignableFrom(clazz) == false) {
                throw new DefaultEngineException("Class "
                        + DefaultEngine.Provider.class
                        + " is not assignable from user defined class "
                        + className
                        + ".");
            }
            Constructor<?> c = null;
            try {
                c = clazz.getConstructor(new Class[] { Document.class });
            } catch (SecurityException se) {
                throw new DefaultEngineException("Error instantiating user defined class "
                        + className
                        + ". "
                        + se.getMessage());
            } catch (NoSuchMethodException ne) {
                throw new DefaultEngineException("User defined class "
                        + className
                        + " must define constructor with single parameter of type "
                        + Document.class.getName()
                        + ".");
            }
            try {
                delegate = (DefaultEngine.Provider)c.newInstance(new Object[] { configuration });
            } catch (Exception ex) {
                throw new DefaultEngineException("Error instantiating user defined class "
                        + className
                        + ". "
                        + ex.getMessage());
            }

            delegate.spi_setConsumer(this.consumer);

            delegate.spi_init(managementContexts);

        }

        public boolean spi_isSuspended() {
            return delegate.spi_isSuspended();
        }

        public void spi_receiveAction(EndpointReference endpointReference, ActionContext actionContext) {
            delegate.spi_receiveAction(endpointReference, actionContext);
        }

        public void spi_resume() throws ExecutableException {
            delegate.spi_resume();
        }

        public void spi_suspend() throws ExecutableException {
            delegate.spi_suspend();
        }

        public void spi_setConsumer(Extension consumer) {
            this.consumer = (DefaultEngine)consumer;
        }

    };

}
