/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.expression.xpath.xpath10;

import java.util.List;

import org.bluestemsoftware.specification.eoa.component.types.rt.ExpressionRT;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactoryContext;
import org.bluestemsoftware.specification.eoa.ext.expression.EvaluationException;
import org.w3c.dom.Node;

/**
 * An <code>Expression</code> object which conforms to version 1.0 of XPath language
 * specification.
 */
public class XPathExpression extends ExpressionRT {

    public static final String TYPE = "http://www.w3.org/TR/xpath";

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ExpressionRT.Provider {
        public void spi_addNamespace(String prefix, String namespace);

        public String spi_evaluateString(Node node) throws EvaluationException;

        public Number spi_evaluateNumber(Node node) throws EvaluationException;

        public Boolean spi_evaluateBoolean(Node node) throws EvaluationException;

        public Node spi_evaluateNode(Node node) throws EvaluationException;

        public List<Node> spi_evaluateNodes(Node node) throws EvaluationException;
    }

    public XPathExpression(XPathExpressionFactory factory, Provider provider) {
        super(factory, provider);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.types.rt.ExpressionRT#addNamespace(java.lang.String,
     *      java.lang.String)
     */
    public void addNamespace(String prefix, String namespace) {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            ExtensionFactoryContext factoryContext = factory.getFactoryContext();
            thread.setContextClassLoader(factoryContext.getClassLoader());
            ((Provider)provider).spi_addNamespace(prefix, namespace);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Evaluates a <code>Node</code> against this <code>Expression</code> according to
     * version 1.0 of XPath language specification and returns result as a <code>String</code>.
     * <p>
     * Note that the underlying implementation of this method MUST be threadsafe, i.e. capable
     * of supporting concurrent use by multiple threads.
     * 
     * @param node
     *        against which this <code>Expression</code> is evaluated.
     * 
     * @return String returns the textual representation of the result of evaluation, i.e. the
     *         'string value' of a given <code>Node</code> type as defined within the XPath
     *         specification.
     * @throws EvaluationException
     *         if an error occurs evaluating expression, e.g. if result of evaluation cannot be
     *         returned as a <code>String</code>.
     */
    public String evaluateString(Node node) throws EvaluationException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            ExtensionFactoryContext factoryContext = factory.getFactoryContext();
            thread.setContextClassLoader(factoryContext.getClassLoader());
            return ((Provider)provider).spi_evaluateString(node);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Evaluates a <code>Node</code> against this <code>Expression</code> according to
     * version 1.0 of XPath language specification and returns result as a <code>Number</code>.
     * <p>
     * Note that the underlying implementation of this method MUST be threadsafe, i.e. capable
     * of supporting concurrent use by multiple threads.
     * 
     * @param node
     *        against which this <code>Expression</code> is evaluated.
     * 
     * @return Number if expression evaluates to a boolean value, the value true is converted
     *         to 1, false is converted to 0. If expression evaluates to a <code>Node</code>
     *         set, the first value in <code>Node</code> set is converted to a String and
     *         then converted to a number according to rules set forth in XPath specification.
     *         If expression evaluates to a String, the String is converted to a number
     *         according to rules set forth in XPath specification. All other datatypes are
     *         converted in a way specific to that type.
     * @throws EvaluationException
     *         if an error occurs evaluating expression, e.g. if result of evaluation cannot be
     *         returned as a <code>Number</code>.
     */
    public Number evaluateNumber(Node node) throws EvaluationException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            ExtensionFactoryContext factoryContext = factory.getFactoryContext();
            thread.setContextClassLoader(factoryContext.getClassLoader());
            return ((Provider)provider).spi_evaluateNumber(node);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Evaluates a <code>Node</code> against this <code>Expression</code> according to
     * version 1.0 of XPath language specification and returns result as a <code>Boolean</code>.
     * <p>
     * Note that the underlying implementation of this method MUST be threadsafe, i.e. capable
     * of supporting concurrent use by multiple threads.
     * 
     * @param node
     *        against which this <code>Expression</code> is evaluated.
     * 
     * @return Boolean a number is true if it is neither positive, negative, zero or NaN - a
     *         <code>NodeSet</code> is true if it is not empty - a String is true if its
     *         length is not zero - all other datatypes are converted in a way specific to that
     *         type.
     * @throws EvaluationException
     *         if an error occurs evaluating expression, e.g. if result of evaluation cannot be
     *         returned as a <code>Boolean</code>.
     */
    public Boolean evaluateBoolean(Node node) throws EvaluationException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            ExtensionFactoryContext factoryContext = factory.getFactoryContext();
            thread.setContextClassLoader(factoryContext.getClassLoader());
            return ((Provider)provider).spi_evaluateBoolean(node);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Evaluates a <code>Node</code> against this <code>Expression</code> according to
     * version 1.0 of XPath language specification and returns result as a <code>Node</code>.
     * <p>
     * Note that the underlying implementation of this method MUST be threadsafe, i.e. capable
     * of supporting concurrent use by multiple threads.
     * 
     * @param node
     *        against which this <code>Expression</code> is evaluated.
     * 
     * @return List<Node> if expression evaluates to multiple <code>Nodes</code>, the first
     *         <code>Node</code> in document order is returned. Note that the returned
     *         <code>Node</code> is not a copy, but is instead a reference to an actual
     *         <code>Node</code> instance within indicated content.
     * @throws EvaluationException
     *         if an error occurs evaluating expression, e.g. if result of evaluation cannot be
     *         returned as a <code>Node</code>.
     */
    public Node evaluateNode(Node node) throws EvaluationException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            ExtensionFactoryContext factoryContext = factory.getFactoryContext();
            thread.setContextClassLoader(factoryContext.getClassLoader());
            return ((Provider)provider).spi_evaluateNode(node);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Evaluates a <code>Node</code> against this <code>Expression</code> according to
     * version 1.0 of XPath language specification and returns result as a <code>Set</code>.
     * <p>
     * Note that the underlying implementation of this method MUST be threadsafe, i.e. capable
     * of supporting concurrent use by multiple threads.
     * 
     * @param node
     *        against which this <code>Expression</code> is evaluated.
     * 
     * @return List<Node> returns a <code>List</code> of matching <code>Nodes</code>.
     *         Returned <code>Nodes</code> are not copies, but are instead references to
     *         actual <code>Node</code> instances within indicated content.
     * @throws EvaluationException
     *         if an error occurs evaluating expression, e.g. if result of evaluation cannot be
     *         returned as a <code>List</code>.
     */
    public List<Node> evaluateNodes(Node node) throws EvaluationException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            ExtensionFactoryContext factoryContext = factory.getFactoryContext();
            thread.setContextClassLoader(factoryContext.getClassLoader());
            return ((Provider)provider).spi_evaluateNodes(node);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

}
