/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.http.client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Collections;
import java.util.Map;

/**
 * Models a request sent by <code>HTTPClient</code>.
 */
public class HTTPClientRequest {

    protected String address;
    protected Method method;
    protected Map<String, String> requestHeaders;
    protected RequestEntity requestEntity;

    public static enum Method {
        GET, POST, PUT, DELETE
    }

    /**
     * Constructs a request with no body and no headers.
     * @param address
     * @param method
     *        which must be either GET or DELETE.
     */
    @SuppressWarnings("unchecked")
    public HTTPClientRequest(String address, Method method) {
        this(address, method, Collections.EMPTY_MAP, null);
    }

    /**
     * Constructs a request.
     * @param address
     * @param method
     *        Indicates method employed. Note that if method is POST or PUT, content type MUST
     *        be defined via a request header.
     * @param requestHeaders
     * @param requestEntity
     *        which MUST NOT be <code>null</code> if method is POST or PUT and which MUST be
     *        null if method is GET or DELETE.
     */
    public HTTPClientRequest(String address, Method method, Map<String, String> requestHeaders,
            RequestEntity requestEntity) {
        if (address == null) {
            throw new IllegalArgumentException("address null");
        } else {
            this.address = address;
        }
        if (method == null) {
            throw new IllegalArgumentException("method null");
        } else {
            this.method = method;
        }
        if (requestHeaders == null) {
            throw new IllegalArgumentException("requestHeaders null");
        } else {
            this.requestHeaders = requestHeaders;
        }
        if (method == Method.POST || method == Method.PUT) {
            if (requestEntity == null) {
                throw new IllegalArgumentException("content null and method is " + method);
            }
        } else {
            if (requestEntity != null) {
                throw new IllegalArgumentException("content provided and method is " + method);
            }
        }
        this.requestEntity = requestEntity;
    }

    /**
     * Gets address to which request is to be sent.
     * 
     * @return
     */
    public String getAddress() {
        return address;
    }

    /**
     * Identifies HTTP method used to send request.
     * @return method
     */
    public Method getMethod() {
        return method;
    }

    /**
     * Gets http request headers to be transmitted with request.
     * @return A map of request headers which maps header name to header value or an empty map
     *         if none defined.
     */
    public Map<String, String> getRequestHeaders() {
        return requestHeaders;
    }

    /**
     * Gets request entity.
     * 
     * @return RequestEntity or <code>null</code> if request has no content, i.e. method is
     *         'GET'.
     */
    public RequestEntity getRequestEntity() throws IOException {
        return requestEntity;
    }

    /**
     * Wraps a request stream.
     */
    public interface RequestEntity {

        /**
         * Indicates whether or not request is empty, i.e. a post/put without any data (e.g.
         * POX over HTTP and conent model is ##none).
         * 
         * @return true if empty, false if not
         */
        public boolean isEmpty();

        /**
         * Writes request entity to output stream.
         * @param out
         * @throws IOException
         */
        public void writeRequest(OutputStream out) throws IOException;

    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#toString()
     * 
     * BUFFERS REQUEST TO MEMORY. SHOULD BE USED FOR DEBUGGING ONLY.
     */
    @Override
    public String toString() {
        try {
            StringBuffer sb = new StringBuffer();
            sb.append("address: ");
            sb.append(address);
            sb.append(System.getProperty("line.separator"));
            sb.append("method: ");
            sb.append(method);
            sb.append(System.getProperty("line.separator"));
            sb.append("headers: ");
            sb.append(System.getProperty("line.separator"));
            sb.append(requestHeaders);
            sb.append(System.getProperty("line.separator"));
            sb.append("content: ");
            sb.append(System.getProperty("line.separator"));
            sb.append(getContentAsString());
            return sb.toString();
        } catch (Exception ex) {
            return super.toString();
        }
    }

    private String getContentAsString() throws IOException {
        if (requestEntity == null) {
            return "";
        }
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        requestEntity.writeRequest(out);
        return new String(out.toByteArray(), "UTF-8");

    }

}
