/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.http.server;

import java.io.File;
import java.util.Map;

import javax.servlet.Servlet;
import javax.servlet.ServletContext;

import org.bluestemsoftware.specification.eoa.ext.feature.FeatureException;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerFeatureException.UniquePathException;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPAuthenticationScheme;
import org.bluestemsoftware.specification.eoa.system.server.Feature;
import org.bluestemsoftware.specification.eoa.system.server.Server;

/**
 * An 'embedded' HTTP Server/Servlet container, i.e. which runs within same jvm as
 * {@link Server}.
 */
public class HTTPServerFeature extends Feature {

    public static final String TYPE = HTTPServerFeature.class.getName();

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends Feature.Provider {

        public int[] spi_getPorts(boolean isSSL);

        public void spi_addServlet(Servlet servlet, String contextPath, ServerAuthInfo authInfo) throws FeatureException;

        public void spi_addServlet(Servlet servlet, String contextPath, ServerAuthInfo authInfo, Map<String, String> parameters, Map<String, Object> attributes) throws FeatureException;

        public void spi_removeServlet(Servlet servlet);

        public void spi_deployWebModule(File webArchive, ClassLoader classLoader, String rootContext) throws HTTPServerFeatureException;

    }

    public HTTPServerFeature(HTTPServerFeatureFactory factory, Feature.Provider provider, File extensionEtcDir, File extensionVarDir) {
        super(factory, provider, extensionEtcDir, extensionVarDir);
    }

    /**
     * Returns port(s) upon which HTTP server is listening for inbound connections of indicated
     * type.
     * 
     * @param isSSL
     *        true if port(s) used for SSL connections, false if not.
     * @return port number(s) or an empty array if no port(s) defined for indicated type
     */
    public int[] getPorts(boolean isSSL) {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_getPorts(isSSL);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Registers <code>Servlet</code> for indicated context, i.e. creates a 'web application'
     * with a single <code>Servlet</code>. Note that is the responsibility of
     * <code>Servlet</code> provider to manage thread context classloader as appropriate.
     * 
     * @param servlet
     *        Servlet which handles all requests for indicated context, i.e. the implied
     *        <code>Servlet</code> path spec is '/*'.
     * @param contextPath
     *        Path against which request URI's are matched, e.g. context path '/mypath' implies
     *        path spec '/mypath/*' and will match any request URI whose path begins with
     *        '/mypath'.
     * 
     * @param authInfo
     *        or <code>null</code> if no authentication required.
     * @throws FeatureException
     * @throws UniquePathException
     *         thrown if pathSpec not unique within context of server.
     */
    public void addServlet(Servlet servlet, String contextPath, ServerAuthInfo authInfo) throws FeatureException {
        if (servlet == null) {
            throw new IllegalArgumentException("servlet null");
        }
        if (contextPath == null) {
            throw new IllegalArgumentException("urlPattern null");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_addServlet(servlet, contextPath, authInfo);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Registers <code>Servlet</code> for indicated context, i.e. creates a 'web application'
     * with a single <code>Servlet</code>. Note that is the responsibility of
     * <code>Servlet</code> provider to manage thread context classloader as appropriate.
     * 
     * @param servlet
     *        Servlet which handles all requests for indicated context, i.e. the implied
     *        <code>Servlet</code> path spec is '/*'.
     * @param contextPath
     *        Path against which request URI's are matched, e.g. context path '/mypath' implies
     *        path spec '/mypath/*' and will match any request URI whose path begins with
     *        '/mypath'.
     * @param authInfo
     *        or <code>null</code> if no authentication required.
     * @param parameters
     *        Init parameters exposed to <code>Servlet</code> via
     *        {@link ServletContext#getInitParameter(String)}
     * @param attributes
     *        Attributes exposed to <code>Servlet</code> via
     *        {@link ServletContext#getAttribute(String)}.
     * @throws FeatureException
     */
    public void addServlet(Servlet servlet, String contextPath, ServerAuthInfo authInfo, Map<String, String> parameters, Map<String, Object> attributes) throws FeatureException {
        if (servlet == null) {
            throw new IllegalArgumentException("servlet null");
        }
        if (contextPath == null) {
            throw new IllegalArgumentException("urlPattern null");
        }
        if (parameters == null) {
            throw new IllegalArgumentException("parameters null");
        }
        if (attributes == null) {
            throw new IllegalArgumentException("attributes null");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_addServlet(servlet, contextPath, authInfo, parameters, attributes);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Removes a 'registered' <code>Servlet</code>.
     * 
     * @param servlet
     */
    public void removeServlet(Servlet servlet) {
        if (servlet == null) {
            throw new IllegalArgumentException("servlet null");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_removeServlet(servlet);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Deploys a <code>WebModule</code>, i.e. which abstracts a j2ee web application.
     * 
     * @param webModule
     * @param classLoader
     *        classloader for <code>Web Application</code> with which module is associated.
     * @param rootContext
     *        of web application
     * @throws HTTPServerFeatureException
     */
    public void deployWebModule(File webArchive, ClassLoader classLoader, String rootContext) throws HTTPServerFeatureException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_deployWebModule(webArchive, classLoader, rootContext);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Encapsulates data used by server to authenticate a client request.
     */
    public interface ServerAuthInfo {

        /**
         * Gets name of realm within which users are defined.
         * @return realm
         */
        public String getRealm();

        /**
         * Gets authentication scheme used by server to authenticate users.
         * @return scheme
         */
        public HTTPAuthenticationScheme getScheme();

    }

}
