/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.jms.server;


import java.io.File;

import javax.jms.ConnectionFactory;
import javax.jms.JMSException;
import javax.jms.Queue;

import org.bluestemsoftware.specification.eoa.system.container.Container;
import org.bluestemsoftware.specification.eoa.system.server.Feature;
import org.bluestemsoftware.specification.eoa.system.server.Server;

/**
 * An 'embedded' JMS Server, i.e. the JMS Server runs within same jvm as {@link Server}.
 */
public class JMSServerFeature extends Feature {

    public static final String TYPE = JMSServerFeature.class.getName();

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends Feature.Provider {
        public ConnectionFactory spi_getEmbeddedConnectionFactory(boolean isXA);

        public int spi_getPort();

        public Queue spi_createQueue(String name) throws JMSException;
    }
    
    public JMSServerFeature(JMSServerFeatureFactory factory, Feature.Provider provider, File extensionEtcDir, File extensionVarDir) {
        super(factory, provider, extensionEtcDir, extensionVarDir);
    }
    
    /**
     * Returns a <code>ConnectionFactory</code> that references an in-memory representation
     * of jms server. Used by <code>ManagedConnectionFactory</code> to create 'managed'
     * connections to embedded server.
     * <p>
     * Note, this method SHOULD NOT be used to create user connections to embedded server.
     * Server SHOULD be accessed via a 'managed' connection configured by container and
     * retreived via {@link Container#getConnector(String)}.
     * 
     * @param isXA
     *        indicates whether or not returned <code>ConnectionFactory</code> creates
     *        connections which are capable of participating in a distributed transaction. If
     *        false, connections supports local transactions only.
     * 
     * @return ConnectionFactory
     */
    public ConnectionFactory getEmbeddedConnectionFactory(boolean isXA) {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_getEmbeddedConnectionFactory(isXA);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Returns port upon which JMS server is listening for inbound connections If server is
     * configured to listen on more than one port, the behavior is undefined.
     * 
     * @return port number or -1 if port undefined
     */
    public int getPort() {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_getPort();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Allows for dynamic creation of a <code>Queue</code>, i.e. as opposed to creating via
     * server configuration.
     * <p>
     * If physical queue does not exist, server MUST create and return it. If physical queue
     * does exist, server MUST return a handle to the existing queue.
     * 
     * 
     * @param name
     *        name of queue
     * @return Queue
     */
    public Queue createQueue(String name) throws JMSException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_createQueue(name);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

}
