/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.InterfaceReference;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.MessagePart;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP11Binding;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP11HTTPProtocol;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP12Binding;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP12HTTPProtocol;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingFaultReference;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingMessageReference;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingOperation;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPDocumentStyleMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPHeaderBlock;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessageExchangePattern;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPRpcStyleMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLFactory.BindingDefinitionReader.BindingStyle;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingOperationElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingOperationFaultElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingOperationInputElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingOperationOutputElement;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyFactory;
import org.bluestemsoftware.specification.eoa.system.SystemContext;

public class SOAPBindingReader {

    public static Binding readBinding(SOAPBindingFactory bindingFactory, ComponentContext componentContext, String targetNamespace, BindingElement bindingElement, BindingStyle bindingStyle) throws DeploymentException {

        ComponentDeployment deployment = componentContext.getDeployment();
        ComponentName bindingName = new ComponentName(targetNamespace, bindingElement.getName());

        // retrieve interface metadata model, which must be
        // dereferenceable

        ComponentName interfaceName = bindingElement.getType();
        Interface intrface = (Interface)deployment.getComponent(ComponentType.INTERFACE, interfaceName);

        // basic profile should have verified its existence, but just in
        // case ...

        InterfaceReference interfaceReference = null;
        if (intrface == null) {
            throw new DeploymentException("Error dereferencing interface "
                    + interfaceName
                    + " bound by binding "
                    + bindingName
                    + ". Definition not found.");
        } else {
            interfaceReference = new InterfaceReference(interfaceName);
        }

        boolean isRPCStyle;
        if (bindingStyle == BindingStyle.RPC_STYLE) {
            isRPCStyle = true;
        } else {
            isRPCStyle = false;
        }

        SOAPBindingElement soapBindingElement = bindingElement.getSOAPBindingElement();

        SOAPVersion soapVersion = null;
        if (soapBindingElement instanceof SOAP11BindingElement) {
            soapVersion = SOAPVersion.SOAP11;
        } else {
            soapVersion = SOAPVersion.SOAP12;
        }

        // we must iterate over operations and create each before creating
        // binding. this is because we must process all faults first.
        // our component model accommodates both wsdl 2.0 and wsdl 1.x.
        // consequently faults are defined at binding level rather than
        // at wsdl 11 operation granularity, i.e. we will not support faults
        // with same name and different message types within a single binding

        Map<QName, SOAPBindingFault> bindingFaults = new HashMap<QName, SOAPBindingFault>();
        Map<QName, SOAPBindingOperation> bindingOperations = new HashMap<QName, SOAPBindingOperation>();

        for (BindingOperationElement bindingOperationElement : bindingElement.getOperationElements().values()) {

            QName operationName = new QName(interfaceName.getNamespaceURI(), bindingOperationElement.getName());
            InterfaceOperation interfaceOperation = intrface.getOperation(operationName);

            // capture the soapAction value, if any. the optional parameterOrder
            // attribute is not relevant as it is used by code generators, e.g.
            // axis wsdl2java tool, to build java objects from an rpc binding and
            // can be safely ignored

            String soapAction = null;
            SOAPOperationElement soapOperation = bindingOperationElement.getSOAPOperationElement();

            if (soapOperation instanceof SOAP11OperationElement) {
                if (soapVersion != SOAPVersion.SOAP11) {
                    throw new DeploymentException("Error importing binding "
                            + bindingName
                            + ". Operation '"
                            + operationName
                            + "' contains both soap11 and soap12 extension elements.");
                }
                soapAction = soapOperation.getSOAPAction();
            } else {
                if (soapVersion != SOAPVersion.SOAP12) {
                    throw new DeploymentException("Error importing binding "
                            + bindingName
                            + ". Operation '"
                            + operationName
                            + "' contains both soap11 and soap12 extension elements.");
                }
                if (((SOAP12OperationElement)soapOperation).isSOAPActionRequired()) {
                    soapAction = soapOperation.getSOAPAction();
                } else {
                    // its optional. don't set on operation, i.e. this is
                    // consistent with WSDL 2.0 rules (which has no such
                    // flag)
                }
            }

            // ************* process binding input **************************

            BindingOperationInputElement bindingInputElement = bindingOperationElement.getInputElement();
            SOAPBodyElement soapBodyElement = bindingInputElement.getSOAPBodyElement();

            if (soapBodyElement instanceof SOAP11BodyElement) {
                if (soapVersion != SOAPVersion.SOAP11) {
                    throw new DeploymentException("Error importing binding "
                            + bindingName
                            + ". Operation '"
                            + operationName
                            + "' contains both soap11 and soap12 extension elements.");
                }
            } else {
                if (soapVersion != SOAPVersion.SOAP12) {
                    throw new DeploymentException("Error importing binding "
                            + bindingName
                            + ". Operation '"
                            + operationName
                            + "' contains both soap11 and soap12 extension elements.");
                }
            }

            List<SOAPHeaderElement> soapHeaderElements = bindingInputElement.getSOAPHeaderElements();

            // create a concrete version of abstract message. abstract message def
            // should have been loaded by MessageHelper. check for null just in case

            InterfaceMessageReference abstractInputMessageRef = interfaceOperation.getInputMessageReference();
            QName abstractInputMessageName = abstractInputMessageRef.getReferencedComponentName();
            InterfaceMessage abstractInputMessage = null;
            abstractInputMessage = (InterfaceMessage)deployment.getComponent(ComponentType.MESSAGE,
                    abstractInputMessageName);
            SOAPMessage soapMessage = getConcreteMessage(soapVersion, bindingName, operationName,
                    abstractInputMessage, soapBodyElement, soapHeaderElements, isRPCStyle);

            // parse element policy attached to message reference element
            ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(bindingInputElement,
                    targetNamespace));

            SOAPBindingMessageReference inputMessageRef = new SOAPBindingMessageReference(Direction.IN,
                    elementPolicy, soapMessage);

            // ************* process binding output **************************

            BindingOperationOutputElement bindingOutputElement = bindingOperationElement.getOutputElement();
            SOAPBindingMessageReference outputMessageRef = null;

            if (bindingOutputElement != null) {

                soapBodyElement = bindingOutputElement.getSOAPBodyElement();

                if (soapBodyElement instanceof SOAP11BodyElement) {
                    if (soapVersion != SOAPVersion.SOAP11) {
                        throw new DeploymentException("Error importing binding "
                                + bindingName
                                + ". Operation '"
                                + operationName
                                + "' contains both soap11 and soap12 extension elements.");
                    }
                } else {
                    if (soapVersion != SOAPVersion.SOAP12) {
                        throw new DeploymentException("Error importing binding "
                                + bindingName
                                + ". Operation '"
                                + operationName
                                + "' contains both soap11 and soap12 extension elements.");
                    }
                }

                soapHeaderElements = bindingOutputElement.getSOAPHeaderElements();

                // create a concrete version of abstract message. abstract message def
                // should have been loaded by MessageHelper. check for null just in case

                InterfaceMessageReference abstractOutputMessageRef = interfaceOperation
                        .getOutputMessageReference();
                QName abstractOutputMessageName = abstractOutputMessageRef.getReferencedComponentName();
                InterfaceMessage abstractOutputMessage = null;
                abstractOutputMessage = (InterfaceMessage)deployment.getComponent(ComponentType.MESSAGE,
                        abstractOutputMessageName);
                soapMessage = getConcreteMessage(soapVersion, bindingName, operationName, abstractOutputMessage,
                        soapBodyElement, soapHeaderElements, isRPCStyle);

                // parse element policy attached to message reference element
                elementPolicy = getAttachedPolicy(componentContext, new DOMSource(bindingOutputElement,
                        targetNamespace));

                outputMessageRef = new SOAPBindingMessageReference(Direction.OUT, elementPolicy, soapMessage);

            }

            // ************* process operation faults **************************

            Map<QName, SOAPBindingFaultReference> faultRefs = new HashMap<QName, SOAPBindingFaultReference>();
            Map<QName, QName> faultMessages = new HashMap<QName, QName>();

            for (BindingOperationFaultElement faultElement : bindingOperationElement.getFaultElements().values()) {

                SOAPFaultElement soapFaultElement = faultElement.getSOAPFaultElement();
                
                if (soapFaultElement instanceof SOAP11FaultElement) {
                    if (soapVersion != SOAPVersion.SOAP11) {
                        throw new DeploymentException("Error importing binding "
                                + bindingName
                                + ". Operation '"
                                + operationName
                                + "' contains both soap11 and soap12 extension elements.");
                    }
                } else {
                    if (soapVersion != SOAPVersion.SOAP12) {
                        throw new DeploymentException("Error importing binding "
                                + bindingName
                                + ". Operation '"
                                + operationName
                                + "' contains both soap11 and soap12 extension elements.");
                    }
                }

                // according to wsdl 2.0 spec, qname of binding fault is same as
                // qname of interface fault to which it is bound. we'll rely on
                // this since wsdl 1.1 fault names are local

                QName faultName = new QName(interfaceName.getNamespaceURI(), faultElement.getName());

                // retrieve abstract fault and associated abstract message. abstract
                // message def should have been loaded by MessageHelper

                InterfaceFaultReference faultRef = interfaceOperation.getFaultReference(faultName);
                QName interfaceFaultMessageName = faultRef.getFault().getReferencedComponentName();
                InterfaceMessage abstractFaultMessage = null;
                abstractFaultMessage = (InterfaceMessage)deployment.getComponent(ComponentType.MESSAGE,
                        interfaceFaultMessageName);

                // generate a concrete version of abstract fault message. fault
                // messages as defined by basic profile are always doc-lit style
                // with a single part (note that wsdl 1.1 schema prohibits the
                // declaration of headers on soap faults, but ... undeclared
                // headers are allowed to occur in message instances. this is
                // not prohibited by basic profile and headers on soap faults
                // are actually required by ws-addressing

                soapMessage = getConcreteMessage(soapVersion, bindingName, operationName, abstractFaultMessage,
                        null, null, false);

                // retrieve element decl which describes soap fault body
                MessagePart ap = abstractFaultMessage.getParts().iterator().next();
                QName faultMessageType = ap.getSchemaComponentQName();

                // no policy attachment to binding fault within wsdl 11 is possible
                elementPolicy = null;

                if (bindingFaults.containsKey(faultName)) {
                    QName existingMessageFactory = faultMessages.get(faultName);
                    if (!existingMessageFactory.equals(faultMessageType)) {
                        throw new DeploymentException("Error importing binding "
                                + bindingName
                                + ". Operation '"
                                + operationName
                                + "' is invalid. At least one other operation throws fault "
                                + faultName
                                + ", but with a different concrete message. This"
                                + " configuration is not supported.");
                    }
                } else {

                    // the soapFaultSubcodes is undefined for soap 1.1 according to
                    // WSDL 20 adjunct which we'll adopt for WSDL 1.1:
                    // http://www.w3.org/TR/wsdl11-soap11-binding/
                    List<QName> soapFaultSubCodes = new ArrayList<QName>();

                    SOAPBindingFault bindingFault = new SOAPBindingFault(faultName, elementPolicy, soapMessage,
                            null, soapFaultSubCodes);
                    bindingFaults.put(faultName, bindingFault);
                    faultMessages.put(faultName, faultMessageType);

                }

                // parse element policy attached to fault reference element
                elementPolicy = getAttachedPolicy(componentContext, new DOMSource(faultElement, targetNamespace));

                SOAPBindingFaultReference bindingFaultRef = new SOAPBindingFaultReference(faultName,
                        Direction.OUT, elementPolicy);
                faultRefs.put(faultName, bindingFaultRef);

            }

            // parse element policy attached to operation element
            elementPolicy = getAttachedPolicy(componentContext, new DOMSource(bindingOperationElement,
                    targetNamespace));

            // the soapMEP is undefined for soap 1.1 according to WSDL 20 adjunct which
            // we'll adopt for WSDL 1.1:
            // http://www.w3.org/TR/wsdl11-soap11-binding/
            SOAPMessageExchangePattern soapMEP = null;

            SOAPBindingOperation concreteOperation = new SOAPBindingOperation(operationName, inputMessageRef,
                    outputMessageRef, faultRefs, elementPolicy, soapMEP, soapAction);
            bindingOperations.put(operationName, concreteOperation);

        }

        // parse element policy attached to binding element
        ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(bindingElement,
                targetNamespace));

        // create the binding

        if (soapVersion == SOAPVersion.SOAP11) {
            return new SOAP11Binding(bindingName, new SOAP11HTTPProtocol(), interfaceReference, elementPolicy,
                    bindingFaults, bindingOperations);
        } else {
            return new SOAP12Binding(bindingName, new SOAP12HTTPProtocol(), interfaceReference, elementPolicy,
                    bindingFaults, bindingOperations, null);
        }

    }

    private static SOAPMessage getConcreteMessage(SOAPVersion soapVersion, QName bindingName, QName operationName, InterfaceMessage abstractMessage, SOAPBodyElement soapBodyElement, List<SOAPHeaderElement> soapHeaderElements, boolean isRPCStyle) throws DeploymentException {

        // parts are abstract within a wsdlMessage. we must therefore determine
        // which part is bound to soap:body by using the soap binding

        List<String> soapBodyParts = new ArrayList<String>();

        String soapBodyNamespace = null;

        if (soapBodyElement != null) {
            
            soapBodyNamespace = soapBodyElement.getNamespace();
            
            List<String> declaredParts = soapBodyElement.getParts();
            
            // process declard parts. note that doc-lit style bindings use parts typed
            // using element attribute, whereas rpc-lit style bindings use parts typed
            // as schema types

            if (declaredParts.size() == 0) {

                // if soapBody defines no parts then abstract message should define zero or
                // one part. zero parts if soap body in input message is empty, 1 part if
                // soap body part is implied. this was already enforced in call to basic
                // profile validator.

                if (abstractMessage.getParts().size() == 1) {
                    MessagePart abstractPart = abstractMessage.getParts().iterator().next();
                    soapBodyParts.add(abstractPart.getName());
                }

            } else {

                // soap body defines one or more parts. doc-lit type bindings may list
                // at most one part. rpc-lit may have unbounded number of parts defined
                // for soap body, ie. these are the parms. these assertions were
                // enforced by profile validator

                for (String part : declaredParts) {
                    soapBodyParts.add(part);
                }

            }
        
        }

        // now we process the optional soap headers of which there may be
        // multiple per message. profile validator enforced that header
        // message ref, if present, is same as input message. see comments
        // on BindingInputValidator

        Map<QName, SOAPHeaderBlock> soapHeaderBlocks = new HashMap<QName, SOAPHeaderBlock>();

        if (soapHeaderElements != null) {
            for (SOAPHeaderElement headerElement : soapHeaderElements) {
                if (headerElement instanceof SOAP11HeaderElement) {
                    if (soapVersion != SOAPVersion.SOAP11) {
                        throw new DeploymentException("Error importing binding "
                                + bindingName
                                + ". Operation '"
                                + operationName
                                + "' contains both soap11 and soap12 extension elements.");
                    }
                } else {
                    if (soapVersion != SOAPVersion.SOAP12) {
                        throw new DeploymentException("Error importing binding "
                                + bindingName
                                + ". Operation '"
                                + operationName
                                + "' contains both soap11 and soap12 extension elements.");
                    }
                }
                String partName = headerElement.getPart();
                MessagePart ap = abstractMessage.getPart(partName);
                QName elementDecl = ap.getSchemaComponentQName();
                SOAPHeaderBlock soapHeaderBlock = new SOAPHeaderBlock(elementDecl, partName);
                if (soapHeaderBlocks.containsKey(soapHeaderBlock.getElementName())) {
                    throw new DeploymentException("Element decl which defines soapHeaderBlock, "
                            + soapHeaderBlock.getElementName()
                            + ", not unique within context of parent component.");
                }
                soapHeaderBlocks.put(soapHeaderBlock.getElementName(), soapHeaderBlock);
            }
        }

        SOAPMessage soapMessage = null;

        // if soap body ns is defined, we know it's an rpc lit message.
        // basic profile validator ensures that ns is not specified if
        // message is doc-lit
        
        if (soapBodyNamespace != null) {
            soapMessage = new SOAPRpcStyleMessage(soapHeaderBlocks, soapBodyNamespace, soapBodyParts);
        } else {
            String soapBodyPart = soapBodyParts.size() == 0 ? null : soapBodyParts.iterator().next();
            soapMessage = new SOAPDocumentStyleMessage(soapHeaderBlocks, soapBodyPart);
        }

        return soapMessage;

    }

    private static ElementPolicy getAttachedPolicy(ComponentContext componentContext, DOMSource source) throws DeploymentException {

        Set<PolicyFactory> factories = null;
        factories = SystemContext.getContext().getSystem().getPolicyFactories();
        ElementPolicy elementPolicy = null;
        for (PolicyFactory factory : factories) {
            ElementPolicy temp = null;
            try {
                temp = factory.readElementPolicy(componentContext, source);
            } catch (DeploymentException de) {
                throw new DeploymentException("Error reading element policy. " + de.getMessage());
            }
            if (elementPolicy == null) {
                elementPolicy = temp;
            } else {
                throw new DeploymentException("Attached policy expressions MUST be identically typed.");
            }
        }
        return elementPolicy;

    }
}
