/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.ext;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.InterfaceReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.ext.binding.http.HTTPBinding;
import org.bluestemsoftware.specification.eoa.ext.binding.http.HTTPBindingFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.http.HTTPBindingFault;
import org.bluestemsoftware.specification.eoa.ext.binding.http.HTTPBindingFaultReference;
import org.bluestemsoftware.specification.eoa.ext.binding.http.HTTPBindingMessageReference;
import org.bluestemsoftware.specification.eoa.ext.binding.http.HTTPBindingOperation;
import org.bluestemsoftware.specification.eoa.ext.binding.http.XMLHTTPHeader;
import org.bluestemsoftware.specification.eoa.ext.binding.http.XMLHTTPProtocol;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPHeader;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPLocationInfo;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPMethod;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLFactory.BindingDefinitionReader.BindingStyle;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingFaultElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingOperationElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingOperationInputElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingOperationOutfaultElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingOperationOutputElement;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyFactory;
import org.bluestemsoftware.specification.eoa.system.SystemContext;

public class HTTPBindingReader {

    public static Binding readBinding(HTTPBindingFactory bindingFactory, ComponentContext componentContext, String targetNamespace, BindingElement bindingElement, BindingStyle bindingStyle) throws DeploymentException {

        HTTPBindingElement httpBindingElement = null;
        if (bindingElement instanceof HTTPBindingElement) {
            httpBindingElement = (HTTPBindingElement)bindingElement;
        } else {
            throw new DeploymentException("Expected binding element to be an instance of "
                    + HTTPBindingElement.class.getName());
        }

        ComponentName bindingName = new ComponentName(targetNamespace, httpBindingElement.getName());

        InterfaceReference interfaceReference = null;
        if (httpBindingElement.getInterface() != null) {
            interfaceReference = new InterfaceReference(httpBindingElement.getInterface());
        }

        Map<QName, Map<String, HTTPHeader>> faultHTTPHeaders = new HashMap<QName, Map<String, HTTPHeader>>();
        Map<QName, Map<Direction, Set<HTTPHeader>>> messageReferenceHTTPHeaders = new HashMap<QName, Map<Direction, Set<HTTPHeader>>>();
        Map<QName, String> operationHTTPContentEncodings = new HashMap<QName, String>();
        Map<QName, String> faultHTTPContentEncodings = new HashMap<QName, String>();
        Map<QName, Integer> errorStatusCodes = new HashMap<QName, Integer>();
        Map<QName, Map<Direction, String>> messageReferenceHTTPContentEncodings = new HashMap<QName, Map<Direction, String>>();
        Map<QName, HTTPLocationInfo> httpLocationInfos = new HashMap<QName, HTTPLocationInfo>();
        Map<QName, String> httpQueryParameterSeparators = new HashMap<QName, String>();
        Map<QName, HTTPMethod> operationMethods = new HashMap<QName, HTTPMethod>();
        Map<QName, String> inputSerializations = new HashMap<QName, String>();
        Map<QName, String> outputSerializations = new HashMap<QName, String>();
        Map<QName, String> faultSerializations = new HashMap<QName, String>();

        // iterate over binding fault elements and create a binding fault
        // component for each

        Map<QName, HTTPBindingFault> bindingFaults = new HashMap<QName, HTTPBindingFault>();
        
        for (BindingFaultElement bindingFaultElement : httpBindingElement.getFaultElements().values()) {

            HTTPBindingFaultElement httpBindingFaultElement = (HTTPBindingFaultElement)bindingFaultElement;

            QName faultName = httpBindingFaultElement.getRef();

            if (bindingFaults.containsKey(faultName)) {
                throw new DeploymentException("Binding fault '" + faultName + "' not unique.");
            }

            if (httpBindingFaultElement.getErrorStatusCode() != null) {
                Integer faultCode = null;
                String temp = httpBindingFaultElement.getErrorStatusCode();
                if (temp.equals("#any")) {
                    faultCode = null;
                } else {
                    faultCode = Integer.valueOf(temp);
                }
                if (faultCode != null) {
                    errorStatusCodes.put(faultName, faultCode);
                }
            }

            // parse optional httpHeader elements each of which defines an
            // http header transmitted with fault message

            for (HTTPHeaderElement headerElement : httpBindingFaultElement.getHTTPHeaderElements().values()) {
                String headerName = headerElement.getName();
                QName headerType = headerElement.getType();
                boolean isRequired = headerElement.isRequired();
                HTTPHeader httpHeader = new XMLHTTPHeader(headerName, headerType, isRequired);
                Map<String, HTTPHeader> faultHeaderMap = faultHTTPHeaders.get(faultName);
                if (faultHeaderMap == null) {
                    faultHeaderMap = new HashMap<String, HTTPHeader>();
                    faultHTTPHeaders.put(faultName, faultHeaderMap);
                }
                faultHeaderMap.put(headerName, httpHeader);
            }

            // parse optional content encoding for fault message
            if (httpBindingFaultElement.getContentEncoding() != null) {
                faultHTTPContentEncodings.put(faultName, httpBindingFaultElement.getContentEncoding());
            }

            // pares element policy attached to binding fault element
            ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(
                    httpBindingFaultElement, targetNamespace));

            // construct the fault
            HTTPBindingFault httpBindingFault = new HTTPBindingFault(faultName, elementPolicy);
            bindingFaults.put(faultName, httpBindingFault);

        }

        Map<QName, HTTPBindingOperation> bindingOperations = new HashMap<QName, HTTPBindingOperation>();

        for (BindingOperationElement operationElement : httpBindingElement.getOperationElements().values()) {

            HTTPBindingOperationElement httpBindingOperationElement = (HTTPBindingOperationElement)operationElement;

            QName operationName = httpBindingOperationElement.getRef();

            // parse optional http method for this operation
            if (httpBindingOperationElement.getMethod() != null) {
                operationMethods.put(operationName, httpBindingOperationElement.getMethod());
            }

            // parse optional content encoding default for this operation
            if (httpBindingOperationElement.getContentEncodingDefault() != null) {
                operationHTTPContentEncodings.put(operationName, httpBindingOperationElement
                        .getContentEncodingDefault());
            }

            // parse optional query parameter separators for this operation
            if (httpBindingOperationElement.getQueryParameterSeparator() != null) {
                httpQueryParameterSeparators.put(operationName, httpBindingOperationElement
                        .getQueryParameterSeparator());
            }

            // parse optional http location attributes for this operation
            if (httpBindingOperationElement.getLocation() != null) {
                String location = httpBindingOperationElement.getLocation();
                Boolean ignoreUncited = httpBindingOperationElement.ignoreUncited();
                httpLocationInfos.put(operationName, new HTTPLocationInfo(operationName, location, ignoreUncited));
            }

            // parse optional input serialization for this method
            if (httpBindingOperationElement.getInputSerialization() != null) {
                inputSerializations.put(operationName, httpBindingOperationElement.getInputSerialization());
            }

            // parse optional output serialization for this method
            if (httpBindingOperationElement.getOutputSerialization() != null) {
                outputSerializations.put(operationName, httpBindingOperationElement.getOutputSerialization());
            }

            // parse optional fault serialization for this method
            if (httpBindingOperationElement.getFaultSerialization() != null) {
                faultSerializations.put(operationName, httpBindingOperationElement.getFaultSerialization());
            }

            // ************* process binding input **************************

            HTTPBindingOperationInputElement inputElement = null;

            for (BindingOperationInputElement temp : httpBindingOperationElement.getInputElements()) {
                if (inputElement == null) {
                    inputElement = (HTTPBindingOperationInputElement)temp;
                } else {
                    throw new DeploymentException("Binding operation "
                            + operationName
                            + " defines more than one input element.");
                }
            }

            HTTPBindingMessageReference inputMessageReference = null;

            if (inputElement != null) {

                // parse optional httpHeader elements each of which defines an
                // http header transmitted with input message

                for (HTTPHeaderElement headerElement : inputElement.getHTTPHeaderElements().values()) {
                    String headerName = headerElement.getName();
                    QName headerType = headerElement.getType();
                    boolean isRequired = headerElement.isRequired();
                    HTTPHeader httpHeader = new XMLHTTPHeader(headerName, headerType, isRequired);
                    Map<Direction, Set<HTTPHeader>> directionMap = messageReferenceHTTPHeaders.get(operationName);
                    if (directionMap == null) {
                        directionMap = new HashMap<Direction, Set<HTTPHeader>>();
                        messageReferenceHTTPHeaders.put(operationName, directionMap);
                    }
                    Set<HTTPHeader> headerSet = directionMap.get(Direction.IN);
                    if (headerSet == null) {
                        headerSet = new HashSet<HTTPHeader>();
                        directionMap.put(Direction.IN, headerSet);
                    }
                    headerSet.add(httpHeader);
                }

                // parse optional content encoding for input message
                if (inputElement.getContentEncoding() != null) {
                    Map<Direction, String> directionMap = messageReferenceHTTPContentEncodings.get(operationName);
                    if (directionMap == null) {
                        directionMap = new HashMap<Direction, String>();
                        messageReferenceHTTPContentEncodings.put(operationName, directionMap);
                    }
                    directionMap.put(Direction.IN, inputElement.getContentEncoding());
                }

                // parses element policy attached to message reference element
                ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(inputElement,
                        targetNamespace));

                inputMessageReference = new HTTPBindingMessageReference(Direction.IN, elementPolicy);

            } else {

                // spec doesn't require one, but we'll create a place holder message ref, i.e.
                // given that all our meps expect an input message ref, we'll default one with
                // no policy so callers don't have to check for null
                
                inputMessageReference = new HTTPBindingMessageReference(Direction.IN, null);

            }
            // ************* process binding output **************************

            HTTPBindingOperationOutputElement outputElement = null;

            for (BindingOperationOutputElement temp : httpBindingOperationElement.getOutputElements()) {
                if (outputElement == null) {
                    outputElement = (HTTPBindingOperationOutputElement)temp;
                } else {
                    throw new DeploymentException("Binding operation "
                            + operationName
                            + " defines more than one output element.");
                }
            }

            HTTPBindingMessageReference outputMessageReference = null;

            if (outputElement != null) {

                // parse optional httpHeader elements each of which defines an
                // http header transmitted with output message

                for (HTTPHeaderElement headerElement : outputElement.getHTTPHeaderElements().values()) {
                    String headerName = headerElement.getName();
                    QName headerType = headerElement.getType();
                    boolean isRequired = headerElement.isRequired();
                    HTTPHeader httpHeader = new XMLHTTPHeader(headerName, headerType, isRequired);
                    Map<Direction, Set<HTTPHeader>> directionMap = messageReferenceHTTPHeaders.get(operationName);
                    if (directionMap == null) {
                        directionMap = new HashMap<Direction, Set<HTTPHeader>>();
                        messageReferenceHTTPHeaders.put(operationName, directionMap);
                    }
                    Set<HTTPHeader> headerSet = directionMap.get(Direction.OUT);
                    if (headerSet == null) {
                        headerSet = new HashSet<HTTPHeader>();
                        directionMap.put(Direction.OUT, headerSet);
                    }
                    headerSet.add(httpHeader);
                }

                // parse optional content encoding for output message
                if (outputElement.getContentEncoding() != null) {
                    Map<Direction, String> directionMap = messageReferenceHTTPContentEncodings.get(operationName);
                    if (directionMap == null) {
                        directionMap = new HashMap<Direction, String>();
                        messageReferenceHTTPContentEncodings.put(operationName, directionMap);
                    }
                    directionMap.put(Direction.OUT, outputElement.getContentEncoding());
                }

                // parses element policy attached to message reference element
                ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(outputElement,
                        targetNamespace));

                outputMessageReference = new HTTPBindingMessageReference(Direction.OUT, elementPolicy);

            }

            // ************* process operation faults **************************

            Map<QName, HTTPBindingFaultReference> faultRefs = new HashMap<QName, HTTPBindingFaultReference>();

            for (BindingOperationOutfaultElement faultRefElement : httpBindingOperationElement
                    .getOutfaultElements()) {

                QName faultName = faultRefElement.getRef();

                // parses element policy attached to fault reference element
                ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(faultRefElement,
                        targetNamespace));

                HTTPBindingFaultReference bindingFaultRef = new HTTPBindingFaultReference(faultName,
                        Direction.OUT, elementPolicy);
                faultRefs.put(faultName, bindingFaultRef);

            }

            if (httpBindingOperationElement.getInfaultElements().size() > 0) {
                throw new DeploymentException("Binding operation "
                        + operationName
                        + " defines at least one infault element. MEP not supported.");
            }

            // parses element policy attached to operation element
            ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(
                    httpBindingOperationElement, targetNamespace));

            HTTPBindingOperation concreteOperation = new HTTPBindingOperation(operationName,
                    inputMessageReference, outputMessageReference, faultRefs, elementPolicy);
            bindingOperations.put(operationName, concreteOperation);

        }

        // parse optional method default defined on binding element
        HTTPMethod methodDefault = httpBindingElement.getMethodDefault();

        // parse optional content encoding default on binding element
        String httpContentEncodingDefault = httpBindingElement.getContentEncodingDefault();

        // parse optional http cookies attrib on binding element
        Boolean useHTTPCookies = httpBindingElement.useHTTPCookies();

        // parse optional default query parameter separator
        String queryParameterSeparatorDefault = httpBindingElement.getQueryParameterSeparatorDefault();

        // parses element policy attached to binding element
        ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(httpBindingElement,
                targetNamespace));

        // create underlying protocol
        XMLHTTPProtocol xmlHTTPProtocol = new XMLHTTPProtocol(methodDefault, useHTTPCookies, faultHTTPHeaders,
                messageReferenceHTTPHeaders, httpContentEncodingDefault, queryParameterSeparatorDefault,
                operationMethods, inputSerializations, outputSerializations, faultSerializations,
                operationHTTPContentEncodings, faultHTTPContentEncodings, errorStatusCodes,
                messageReferenceHTTPContentEncodings, httpLocationInfos, httpQueryParameterSeparators);

        Binding binding = new HTTPBinding(bindingName, xmlHTTPProtocol, interfaceReference, elementPolicy,
                bindingFaults, bindingOperations);

        return binding;
    }

    private static ElementPolicy getAttachedPolicy(ComponentContext componentContext, DOMSource source) throws DeploymentException {

        Set<PolicyFactory> factories = null;
        factories = SystemContext.getContext().getSystem().getPolicyFactories();
        ElementPolicy elementPolicy = null;
        for (PolicyFactory factory : factories) {
            ElementPolicy temp = null;
            try {
                temp = factory.readElementPolicy(componentContext, source);
            } catch (DeploymentException de) {
                throw new DeploymentException("Error reading element policy. " + de.getMessage());
            }
            if (elementPolicy == null) {
                elementPolicy = temp;
            } else {
                throw new DeploymentException("Attached policy expressions MUST be identically typed.");
            }
        }
        return elementPolicy;

    }

}
