/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.ext;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.InterfaceReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.message.MessagePart;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP11Binding;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP11HTTPProtocol;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP11VMTransportProtocol;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP12Binding;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP12HTTPProtocol;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAP12VMTransportProtocol;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingFaultReference;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingMessageReference;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingOperation;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPDocumentStyleMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPHTTPHeader;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPHeaderBlock;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessageExchangePattern;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPUnderlyingProtocol;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPHeader;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPLocationInfo;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLFactory.BindingDefinitionReader.BindingStyle;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingFaultElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingOperationElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingOperationInputElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingOperationOutfaultElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingOperationOutputElement;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyFactory;
import org.bluestemsoftware.specification.eoa.system.SystemContext;

public class SOAPBindingReader {

    public static Binding readBinding(SOAPBindingFactory bindingFactory, ComponentContext componentContext, String targetNamespace, BindingElement bindingElement, BindingStyle bindingStyle) throws DeploymentException {

        SOAPBindingElement soapBindingElement = null;
        if (bindingElement instanceof SOAPBindingElement) {
            soapBindingElement = (SOAPBindingElement)bindingElement;
        } else {
            throw new DeploymentException("Expected binding element to be an instance of "
                    + SOAPBindingElement.class.getName());
        }

        ComponentName bindingName = new ComponentName(targetNamespace, soapBindingElement.getName());

        InterfaceReference interfaceReference = null;
        if (soapBindingElement.getInterface() != null) {
            interfaceReference = new InterfaceReference(soapBindingElement.getInterface());
        }

        // parse soap version employed by this binding
        SOAPVersion soapVersion = null;
        if (soapBindingElement.getVersion() == null) {
            soapVersion = SOAPVersion.SOAP12;
        } else {
            if (soapBindingElement.getVersion().equals("1.1")) {
                soapVersion = SOAPVersion.SOAP11;
            } else if (soapBindingElement.getVersion().equals("1.2")) {
                soapVersion = SOAPVersion.SOAP12;
            } else {
                throw new DeploymentException("wsoap:version specifies an unspported value '"
                        + soapBindingElement.getVersion()
                        + "'.");
            }
        }

        // parse default soap mep. if soap version is 1.1, then default mep should be
        // undefined
        SOAPMessageExchangePattern defaultSOAPMep = null;
        if (soapBindingElement.getMEPDefault() != null) {
            defaultSOAPMep = SOAPMessageExchangePattern.valueOfURI(soapBindingElement.getMEPDefault());
        }

        if (defaultSOAPMep != null && soapVersion == SOAPVersion.SOAP11) {
            throw new DeploymentException("SOAP 1.1. binding must not define a value for wsoap:mepDefault");
        }

        // set up collections which will be used to create a soap 1.2
        // over http transport. if soap version is 1.1, the collections
        // are unused

        Map<QName, Map<String, HTTPHeader>> faultHTTPHeaders = new HashMap<QName, Map<String, HTTPHeader>>();
        Map<QName, Map<Direction, Set<HTTPHeader>>> messageReferenceHTTPHeaders = new HashMap<QName, Map<Direction, Set<HTTPHeader>>>();
        Map<QName, String> operationHTTPContentEncodings = new HashMap<QName, String>();
        Map<QName, String> faultHTTPContentEncodings = new HashMap<QName, String>();
        Map<QName, Map<Direction, String>> messageReferenceHTTPContentEncodings = new HashMap<QName, Map<Direction, String>>();
        Map<QName, HTTPLocationInfo> httpLocationInfos = new HashMap<QName, HTTPLocationInfo>();
        Map<QName, String> httpQueryParameterSeparators = new HashMap<QName, String>();

        // iterate over binding fault elements and create a binding fault
        // component for each

        Map<QName, SOAPBindingFault> bindingFaults = new HashMap<QName, SOAPBindingFault>();

        for (BindingFaultElement bindingFaultElement : soapBindingElement.getFaultElements().values()) {

            SOAPBindingFaultElement soapBindingFaultElement = (SOAPBindingFaultElement)bindingFaultElement;

            QName faultName = soapBindingFaultElement.getRef();

            if (bindingFaults.containsKey(faultName)) {
                throw new DeploymentException("Bindng fault '" + faultName + "' not unique.");
            }

            QName soapFaultCode = null;
            if (soapBindingFaultElement.getCode() != null) {
                String temp = soapBindingFaultElement.getCode();
                if (temp.equals("#any")) {
                    soapFaultCode = null;
                } else {
                    soapFaultCode = ComponentName.valueOf(soapBindingFaultElement, temp);
                }
            }

            List<QName> soapFaultSubCodes = new ArrayList<QName>();
            if (soapBindingFaultElement.getSubcodes() != null) {
                StringTokenizer st = new StringTokenizer(soapBindingFaultElement.getSubcodes());
                while (st.hasMoreTokens()) {
                    soapFaultSubCodes.add(ComponentName.valueOf(soapBindingFaultElement, st.nextToken()));
                }
            }

            // parse optional httpHeader elements each of which defines an
            // http header transmitted with fault message

            for (HTTPHeaderElement headerElement : soapBindingFaultElement.getHTTPHeaderElements().values()) {
                String headerName = headerElement.getName();
                QName headerType = headerElement.getType();
                boolean isRequired = headerElement.isRequired();
                HTTPHeader httpHeader = new SOAPHTTPHeader(headerName, headerType, isRequired);
                Map<String, HTTPHeader> faultHeaderMap = faultHTTPHeaders.get(faultName);
                if (faultHeaderMap == null) {
                    faultHeaderMap = new HashMap<String, HTTPHeader>();
                    faultHTTPHeaders.put(faultName, faultHeaderMap);
                }
                faultHeaderMap.put(headerName, httpHeader);
            }

            // parse optional content encoding for fault message
            if (soapBindingFaultElement.getContentEncoding() != null) {
                faultHTTPContentEncodings.put(faultName, soapBindingFaultElement.getContentEncoding());
            }

            // pares element policy attached to binding fault element
            ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(
                    soapBindingFaultElement, targetNamespace));

            // parse concrete message details, i.e. payload content
            // and optional soap headers
            Map<QName, SOAPHeaderBlock> soapHeaderBlocks = new HashMap<QName, SOAPHeaderBlock>();
            for (SOAPHeaderElement headerElement : soapBindingFaultElement.getSOAPHeaderElements().values()) {
                SOAPHeaderBlock soapHeaderBlock = createSOAPHeader(headerElement);
                if (soapHeaderBlocks.containsKey(soapHeaderBlock.getElementName())) {
                    throw new DeploymentException("Element decl which defines soapHeaderBlock, "
                            + soapHeaderBlock.getElementName()
                            + ", not unique within context of parent component.");
                }
                soapHeaderBlocks.put(soapHeaderBlock.getElementName(), soapHeaderBlock);
            }

            SOAPMessage concreteFaultMessage = new SOAPDocumentStyleMessage(soapHeaderBlocks, MessagePart.PAYLOAD);

            // construct the fault
            SOAPBindingFault soapBindingFault = new SOAPBindingFault(faultName, elementPolicy,
                    concreteFaultMessage, soapFaultCode, soapFaultSubCodes);
            bindingFaults.put(faultName, soapBindingFault);

        }

        Map<QName, SOAPBindingOperation> bindingOperations = new HashMap<QName, SOAPBindingOperation>();

        for (BindingOperationElement operationElement : soapBindingElement.getOperationElements().values()) {

            SOAPBindingOperationElement soapBindingOperationElement = (SOAPBindingOperationElement)operationElement;

            QName operationName = soapBindingOperationElement.getRef();

            SOAPMessageExchangePattern soapMEP = null;
            if (soapBindingOperationElement.getMEP() != null) {
                soapMEP = SOAPMessageExchangePattern.valueOfURI(soapBindingOperationElement.getMEP());
            }

            if (soapMEP != null && soapVersion == SOAPVersion.SOAP11) {
                throw new DeploymentException("Operations defined on a SOAP 1.1. binding must not"
                        + " define a value for wsoap:mep.");
            }

            // parse soap action value
            String soapAction = null;
            if (soapBindingOperationElement.getAction() != null) {
                soapAction = soapBindingOperationElement.getAction();
            }

            // parse optional content encoding default for this operation
            if (soapBindingOperationElement.getContentEncodingDefault() != null) {
                operationHTTPContentEncodings.put(operationName, soapBindingOperationElement
                        .getContentEncodingDefault());
            }

            // parse optional query parameter separators for this operation
            if (soapBindingOperationElement.getQueryParameterSeparator() != null) {
                httpQueryParameterSeparators.put(operationName, soapBindingOperationElement
                        .getQueryParameterSeparator());
            }

            // parse optional http location attributes for this operation
            if (soapBindingOperationElement.getLocation() != null) {
                String location = soapBindingOperationElement.getLocation();
                Boolean ignoreUncited = soapBindingOperationElement.ignoreUncited();
                httpLocationInfos.put(operationName, new HTTPLocationInfo(operationName, location, ignoreUncited));
            }

            // ************* process binding input **************************

            SOAPBindingOperationInputElement inputElement = null;

            for (BindingOperationInputElement temp : soapBindingOperationElement.getInputElements()) {
                if (inputElement == null) {
                    inputElement = (SOAPBindingOperationInputElement)temp;
                } else {
                    throw new DeploymentException("Binding operation "
                            + operationName
                            + " defines more than one input element.");
                }
            }

            SOAPBindingMessageReference inputMessageReference = null;
            
            if (inputElement != null) {

                // parse optional httpHeader elements each of which defines an
                // http header transmitted with input message

                for (HTTPHeaderElement headerElement : inputElement.getHTTPHeaderElements().values()) {
                    String headerName = headerElement.getName();
                    QName headerType = headerElement.getType();
                    boolean isRequired = headerElement.isRequired();
                    HTTPHeader httpHeader = new SOAPHTTPHeader(headerName, headerType, isRequired);
                    Map<Direction, Set<HTTPHeader>> directionMap = messageReferenceHTTPHeaders.get(operationName);
                    if (directionMap == null) {
                        directionMap = new HashMap<Direction, Set<HTTPHeader>>();
                        messageReferenceHTTPHeaders.put(operationName, directionMap);
                    }
                    Set<HTTPHeader> headerSet = directionMap.get(Direction.IN);
                    if (headerSet == null) {
                        headerSet = new HashSet<HTTPHeader>();
                        directionMap.put(Direction.IN, headerSet);
                    }
                    headerSet.add(httpHeader);
                }

                // parse optional content encoding for input message
                if (inputElement.getContentEncoding() != null) {
                    Map<Direction, String> directionMap = messageReferenceHTTPContentEncodings.get(operationName);
                    if (directionMap == null) {
                        directionMap = new HashMap<Direction, String>();
                        messageReferenceHTTPContentEncodings.put(operationName, directionMap);
                    }
                    directionMap.put(Direction.IN, inputElement.getContentEncoding());
                }

                // create a concrete version of abstract input message
                SOAPMessage inputMessage = null;

                Map<QName, SOAPHeaderBlock> soapHeaderBlocks = new HashMap<QName, SOAPHeaderBlock>();
                for (SOAPHeaderElement headerElement : inputElement.getSOAPHeaderElements().values()) {
                    SOAPHeaderBlock soapHeaderBlock = createSOAPHeader(headerElement);
                    if (soapHeaderBlocks.containsKey(soapHeaderBlock.getElementName())) {
                        throw new DeploymentException("Element decl which defines soapHeaderBlock, "
                                + soapHeaderBlock.getElementName()
                                + ", not unique within context of parent component.");
                    }
                    soapHeaderBlocks.put(soapHeaderBlock.getElementName(), soapHeaderBlock);
                }

                inputMessage = new SOAPDocumentStyleMessage(soapHeaderBlocks, MessagePart.PAYLOAD);

                // parses element policy attached to message reference element
                ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(inputElement,
                        targetNamespace));

                inputMessageReference = new SOAPBindingMessageReference(Direction.IN, elementPolicy, inputMessage);

            } else {

                // spec doesn't require one, but we'll create a place holder message ref with
                // implied values, i.e. given that all our meps expect an input message ref
                // so callers don't have to check for null
                
                Map<QName, SOAPHeaderBlock> soapHeaderBlocks = new HashMap<QName, SOAPHeaderBlock>();
                SOAPMessage inputMessage = new SOAPDocumentStyleMessage(soapHeaderBlocks, MessagePart.PAYLOAD);
                inputMessageReference = new SOAPBindingMessageReference(Direction.IN, null, inputMessage);
                
            }

            // ************* process binding output **************************

            SOAPBindingOperationOutputElement outputElement = null;

            for (BindingOperationOutputElement temp : soapBindingOperationElement.getOutputElements()) {
                if (outputElement == null) {
                    outputElement = (SOAPBindingOperationOutputElement)temp;
                } else {
                    throw new DeploymentException("Binding operation "
                            + operationName
                            + " defines more than one output element.");
                }
            }

            SOAPBindingMessageReference outputMessageReference = null;

            if (outputElement != null) {

                // parse optional httpHeader elements each of which defines an
                // http header transmitted with output message

                for (HTTPHeaderElement headerElement : outputElement.getHTTPHeaderElements().values()) {
                    String headerName = headerElement.getName();
                    QName headerType = headerElement.getType();
                    boolean isRequired = headerElement.isRequired();
                    HTTPHeader httpHeader = new SOAPHTTPHeader(headerName, headerType, isRequired);
                    Map<Direction, Set<HTTPHeader>> directionMap = messageReferenceHTTPHeaders.get(operationName);
                    if (directionMap == null) {
                        directionMap = new HashMap<Direction, Set<HTTPHeader>>();
                        messageReferenceHTTPHeaders.put(operationName, directionMap);
                    }
                    Set<HTTPHeader> headerSet = directionMap.get(Direction.OUT);
                    if (headerSet == null) {
                        headerSet = new HashSet<HTTPHeader>();
                        directionMap.put(Direction.OUT, headerSet);
                    }
                    headerSet.add(httpHeader);
                }

                // parse optional content encoding for output message
                if (outputElement.getContentEncoding() != null) {
                    Map<Direction, String> directionMap = messageReferenceHTTPContentEncodings.get(operationName);
                    if (directionMap == null) {
                        directionMap = new HashMap<Direction, String>();
                        messageReferenceHTTPContentEncodings.put(operationName, directionMap);
                    }
                    directionMap.put(Direction.OUT, outputElement.getContentEncoding());
                }

                // create a concrete version of abstract output message
                SOAPMessage outputMessage = null;

                Map<QName, SOAPHeaderBlock> soapHeaderBlocks = new HashMap<QName, SOAPHeaderBlock>();
                for (SOAPHeaderElement headerElement : outputElement.getSOAPHeaderElements().values()) {
                    SOAPHeaderBlock soapHeaderBlock = createSOAPHeader(headerElement);
                    if (soapHeaderBlocks.containsKey(soapHeaderBlock.getElementName())) {
                        throw new DeploymentException("Element decl which defines soapHeaderBlock, "
                                + soapHeaderBlock.getElementName()
                                + ", not unique within context of parent component.");
                    }
                    soapHeaderBlocks.put(soapHeaderBlock.getElementName(), soapHeaderBlock);
                }

                outputMessage = new SOAPDocumentStyleMessage(soapHeaderBlocks, MessagePart.PAYLOAD);

                // parses element policy attached to message reference element
                ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(outputElement, targetNamespace));

                outputMessageReference = new SOAPBindingMessageReference(Direction.OUT, elementPolicy,
                        outputMessage);

            }

            // ************* process operation faults **************************

            Map<QName, SOAPBindingFaultReference> faultRefs = new HashMap<QName, SOAPBindingFaultReference>();

            for (BindingOperationOutfaultElement faultRefElement : soapBindingOperationElement
                    .getOutfaultElements()) {

                QName faultName = faultRefElement.getRef();

                // parses element policy attached to fault reference element
                ElementPolicy elementPolicy = getAttachedPolicy(componentContext,
                        new DOMSource(faultRefElement, targetNamespace));

                SOAPBindingFaultReference bindingFaultRef = new SOAPBindingFaultReference(faultName,
                        Direction.OUT, elementPolicy);
                faultRefs.put(faultName, bindingFaultRef);

            }

            if (soapBindingOperationElement.getInfaultElements().size() > 0) {
                throw new DeploymentException("Binding operation "
                        + operationName
                        + " defines at least one infault element. MEP not supported.");
            }

            // parses element policy attached to operation element
            ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(soapBindingOperationElement,
                    targetNamespace));

            SOAPBindingOperation concreteOperation = new SOAPBindingOperation(operationName,
                    inputMessageReference, outputMessageReference, faultRefs, elementPolicy, soapMEP, soapAction);
            bindingOperations.put(operationName, concreteOperation);

        }

        // parse optional content encoding default on binding element
        String httpContentEncodingDefault = null;
        if (soapBindingElement.getContentEncodingDefault() != null) {
            httpContentEncodingDefault = soapBindingElement.getContentEncodingDefault();
        }

        // parse optional http cookies attrib on binding element
        Boolean useHTTPCookies = soapBindingElement.useHTTPCookies();

        // parse optional default query parameter separator
        String queryParameterSeparatorDefault = soapBindingElement.getQueryParameterSeparatorDefault();

        // parse soap underlying protocol into an underlying protocol
        // component

        SOAPUnderlyingProtocol soapUnderlyingProtocol = null;
        if (soapBindingElement.getProtocol() != null) {
            String temp = soapBindingElement.getProtocol();
            if (temp.equals(SOAP11HTTPProtocol.TYPE)) {
                soapUnderlyingProtocol = new SOAP11HTTPProtocol();
            } else if (temp.equals(SOAP12HTTPProtocol.TYPE)) {
                soapUnderlyingProtocol = new SOAP12HTTPProtocol(useHTTPCookies, faultHTTPHeaders,
                        messageReferenceHTTPHeaders, httpContentEncodingDefault, queryParameterSeparatorDefault,
                        operationHTTPContentEncodings, faultHTTPContentEncodings,
                        messageReferenceHTTPContentEncodings, httpLocationInfos, httpQueryParameterSeparators);
            } else if (temp.equals(SOAP11VMTransportProtocol.TYPE)) {
                soapUnderlyingProtocol = new SOAP11VMTransportProtocol();
            } else if (temp.equals(SOAP12VMTransportProtocol.TYPE)) {
                soapUnderlyingProtocol = new SOAP12VMTransportProtocol();
            } else {
                throw new DeploymentException("SOAP binding specifies unsupported value '"
                        + temp
                        + "' for wsoap:protocol.");
            }
        } else {
            throw new DeploymentException("SOAP binding fails to define a value for wsoap:protocol.");
        }

        // pares element policy attached to binding element
        ElementPolicy elementPolicy = getAttachedPolicy(componentContext, new DOMSource(soapBindingElement,
                targetNamespace));

        Binding binding = null;
        if (soapVersion == SOAPVersion.SOAP11) {
            binding = new SOAP11Binding(bindingName, soapUnderlyingProtocol, interfaceReference, elementPolicy,
                    bindingFaults, bindingOperations);
        } else {
            binding = new SOAP12Binding(bindingName, soapUnderlyingProtocol, interfaceReference, elementPolicy,
                    bindingFaults, bindingOperations, defaultSOAPMep);
        }

        return binding;
    }

    private static SOAPHeaderBlock createSOAPHeader(SOAPHeaderElement soapHeaderElement) throws DeploymentException {

        QName elementName = soapHeaderElement.getElement();

        if (elementName == null) {
            throw new DeploymentException("Failed to resolve element reference on wsoap:header."
                    + " Missing required 'element' attribute.");
        }

        return new SOAPHeaderBlock(elementName, soapHeaderElement.mustUnderstand(), soapHeaderElement.isRequired());

    }

    private static ElementPolicy getAttachedPolicy(ComponentContext componentContext, DOMSource source) throws DeploymentException {

        Set<PolicyFactory> factories = null;
        factories = SystemContext.getContext().getSystem().getPolicyFactories();
        ElementPolicy elementPolicy = null;
        for (PolicyFactory factory : factories) {
            ElementPolicy temp = null;
            try {
                temp = factory.readElementPolicy(componentContext, source);
            } catch (DeploymentException de) {
                throw new DeploymentException("Error reading element policy. " + de.getMessage());
            }
            if (elementPolicy == null) {
                elementPolicy = temp;
            } else {
                throw new DeploymentException("Attached policy expressions MUST be identically typed.");
            }
        }
        return elementPolicy;

    }

}
