/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.management.portlet;

import java.io.File;
import java.net.URL;

import org.bluestemsoftware.specification.eoa.DeploymentClassLoader;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactoryContext;
import org.bluestemsoftware.specification.eoa.ext.ManageableExtension;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerFeatureException;
import org.bluestemsoftware.specification.eoa.ext.management.ManagementContextFactory;
import org.bluestemsoftware.specification.eoa.system.ManagementContext;

/**
 * A type specific instance of <code>ManagementContext</code>, i.e. on which provides ability
 * to manage system and deployed extensions via a portal.
 */
public class PortletContainerContext extends ManagementContext {

    public static final String TYPE = PortletContainerContext.class.getName();
  
    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ManagementContext.Provider {
        public void spi_deployPortlet(File webArchive, ManageableExtension manageableExtension, PortletClassLoader portletClassLoader, String portletName, String rootContext) throws PortletContainerContextException;
        public void spi_addListener(PortletContainerListener portletContainerListener);
    }

    public PortletContainerContext(ManagementContextFactory factory, ManagementContext.Provider provider, File extensionEtcDir, File extensionVarDir) {
        super(factory, provider, extensionEtcDir, extensionVarDir);
    }

    /**
     * Add an interested listener. Registered listeners will be notified following the
     * completion of each event as modeled by {@link PortletContainerEvent}.
     * @param portletContainerListener
     */
    public void addListener(PortletContainerListener portletContainerListener) {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_addListener(portletContainerListener);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Deploys a <code>Portlet</code> which serves to manage a {@link ManageagleExtension}
     * within the context of a portal.
     * 
     * @param webArchive
     *        portlet war
     * @param manageableExtension
     *        which defines the portlet as a dependency
     * @param portletName
     *        as defined within portlet descriptor (portlet.xml.) Note that this specification
     *        supports deployment of only one portlet per 'portlet application,' i.e. each
     *        manageable extension MUST NOT define more than one portlet.
     * 
     * @throws HTTPServerFeatureException
     */
    public void deployPortlet(File webArchive, ManageableExtension manageableExtension, String portletName) throws PortletContainerContextException {

        // create the portlet classloader

        DeploymentClassLoader dcl = factory.getFactoryContext().getClassLoader();
        ExtensionFactoryContext ctxt = manageableExtension.getExtensionFactory().getFactoryContext();
        DeploymentClassLoader mel = ctxt.getClassLoader();
        String id = manageableExtension.getExtensionFactory().getName();
        PortletClassLoader pcl = new PortletClassLoader(id + "-portlet", new URL[] {}, dcl, mel);

        // derive portlet root context, which must be unique within context
        // of http server. because only one version of portlet dependency is
        // allowed and because all are typed as war, organization and
        // artifactID will suffice

        String oid = ctxt.getDeployment().getOrganizationID();
        String aid = ctxt.getDeployment().getArtifactID();
        String rootContext = oid + "." + aid;

        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_deployPortlet(webArchive, manageableExtension, pcl, portletName, rootContext);
        } finally {
            thread.setContextClassLoader(cl);
        }

    }

}
