/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.server.standalone10;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.Resource;
import org.bluestemsoftware.specification.eoa.ext.server.ServerFactory;
import org.bluestemsoftware.specification.eoa.system.server.EnabledFeature;
import org.bluestemsoftware.specification.eoa.system.server.Server;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * A factory for producing a 'standalone' <code>Server</code>.
 */

public final class StandaloneServerFactory extends ServerFactory {

    public static final String NAME = "standalone-server";
    public static final String SYSTEM_REF = "http://bluestemsoftware.org/open/eoa/system/standalone";

    public interface Provider extends ServerFactory.Provider {
        public Server spi_createServer(StandaloneServerFactory factory, Map<String, EnabledFeature> enabledFeatures, File extensionEtcDir, File extensionVarDir) throws StandaloneServerException; 
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExtensionFactory#getExtensionType()
     */
    public final String getExtensionType() {
        return StandaloneServer.TYPE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.server.ServerFactory#createServer(javax.xml.transform.dom.DOMSource)
     */
    public Server createServer(DOMSource domSource) throws StandaloneServerException {

        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factoryContext.getClassLoader());
            // validate the definition and parse server name, which we use to version
            // configuration

            validateDefinition(domSource);
            Element definition = (Element)domSource.getNode();

            // parse managed features, i.e. a forced implementation along with an
            // optional configuration which overrides default provided by factory.
            // note that we use a linked hashmap such that we can get predictable
            // order for feature enablement within our tests

            Map<String, EnabledFeature> enabledFeatures = new LinkedHashMap<String, EnabledFeature>();
            QName childName = new QName(StandaloneServer.TYPE, "features");
            Element featuresElement = getChildElement(definition, childName);
            if (featuresElement != null) {
                List<Element> featureElements = getChildElements(featuresElement, new QName(
                        StandaloneServer.TYPE, "feature"));
                for (Element featureElement : featureElements) {
                    String featureType = featureElement.getAttribute("type");
                    String impl = featureElement.getAttribute("impl");
                    impl = impl.equals("") ? null : impl;
                    if (!enabledFeatures.containsKey(featureType)) {
                        enabledFeatures.put(featureType, new EnabledFeature(featureType, impl));
                    } else {
                        throw new StandaloneServerException("Failed to parse configuration for Server."
                                + " More than one Feature of type '"
                                + featureType
                                + " specified.");
                    }
                }
            }
            File extensionEtcDir = createServerEtcDir();
            File extensionVarDir = createServerVarDir();
            return ((Provider)provider).spi_createServer(this, enabledFeatures, extensionEtcDir, extensionVarDir);

        } finally {
            thread.setContextClassLoader(cl);
        }

    }

   private List<Element> getChildElements(Element parent, QName childName) {
        List<Element> answer = new ArrayList<Element>();
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                if (element.getNamespaceURI() != null) {
                    continue;
                }
            } else {
                if (element.getNamespaceURI() == null) {
                    continue;
                }
                if (!childName.getNamespaceURI().equals("*")
                        && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                    continue;
                }
            }
            if (!childName.getLocalPart().equals("*") && !element.getLocalName().equals(childName.getLocalPart())) {
                continue;
            }
            answer.add(element);
        }
        return answer;
    }

    private Element getChildElement(Element parent, QName childName) {
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                if (element.getNamespaceURI() != null) {
                    continue;
                }
            } else {
                if (element.getNamespaceURI() == null) {
                    continue;
                }
                if (!childName.getNamespaceURI().equals("*")
                        && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                    continue;
                }
            }
            if (!childName.getLocalPart().equals("*") && !element.getLocalName().equals(childName.getLocalPart())) {
                continue;
            }
            return element;
        }
        return null;
    }

    private void validateDefinition(DOMSource source) throws StandaloneServerException {

        Resource resource = null;
        try {
            resource = factoryContext.getResource(StandaloneServer.PROVIDER_SCHEMA_LOCATION);
        } catch (DeploymentException de) {
            throw new StandaloneServerException("Error validating definition. " + de);
        }

        SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

        javax.xml.validation.Schema schema = null;
        try {
            schema = schemaFactory.newSchema(new StreamSource(resource.getInputStream()));
        } catch (Exception ex) {
            throw new StandaloneServerException("Error parsing schema. " + ex);
        }
        DOMValidator domValidator = new DOMValidator(schema);
        String validationError;
        try {
            validationError = domValidator.validate(source);
        } catch (Exception ex) {
            throw new StandaloneServerException("Error validating definition. " + ex);
        }
        if (validationError != null) {
            throw new StandaloneServerException("Engine definition invalid. " + validationError);
        }

    }

    static final class DOMValidator {

        private Schema schema;

        public DOMValidator(Schema schema) {
            this.schema = schema;
        }

        public String validate(DOMSource source) throws SAXException, IOException {
            Validator validator = schema.newValidator();
            validator.setErrorHandler(new ErrorHandlerImpl());
            validator.validate(source);
            return ((ErrorHandlerImpl)validator.getErrorHandler()).getErrors();
        }

        static class ErrorHandlerImpl implements ErrorHandler {

            private ArrayList<String> errorMessages = new ArrayList<String>();

            public void error(SAXParseException spe) throws SAXException {
                errorMessages.add(spe.getMessage());
            }

            public void warning(SAXParseException spe) throws SAXException {
                errorMessages.add(spe.getMessage());
            }

            public void fatalError(SAXParseException spe) throws SAXException {
                errorMessages.add(spe.getMessage());
            }

            public String getErrors() {
                if (errorMessages.size() > 0) {
                    Iterator<String> messageIterator = errorMessages.iterator();
                    StringBuilder errorMessage = new StringBuilder();
                    while (messageIterator.hasNext()) {
                        errorMessage.append(messageIterator.next() + "  ");
                    }
                    return errorMessage.toString();
                }
                return null;
            }

        }

    }
    
}
